import copy
import weakref
from typing import Any, Callable, Dict, List, Optional, Tuple

import jax
import jax.numpy as jnp
import torch
import torch.nn as nn
import torch.utils.dlpack
from brax import envs
from brax.io import html, image
from evox.core import Problem, use_state
from evox.problems.neuroevolution.utils import get_vmap_model_state_forward
from evox.utils import VmapInfo, clamp
from torch._C._functorch import get_unwrapped, is_batchedtensor


# to_dlpack is not necessary for torch.Tensor and jax.Array
# because they have a __dlpack__ method, which is called by their respective from_dlpack methods.
def to_jax_array(x: torch.Tensor) -> jax.Array:
    # When the torch has GPU support but the jax does not, we need to move the tensor to CPU first.
    if is_batchedtensor(x):
        x = get_unwrapped(x)
    if x.device.type != "cpu" and jax.default_backend() == "cpu":
        return jax.dlpack.from_dlpack(x.detach().cpu())
    return jax.dlpack.from_dlpack(x.detach())


def from_jax_array(x: jax.Array, device: Optional[torch.device] = None) -> torch.Tensor:
    if device is None:
        device = torch.get_default_device()
    return torch.utils.dlpack.from_dlpack(x).to(device)


__brax_data__: Dict[
    int,
    Tuple[
        Callable[[jax.Array], envs.State],  # vmap_brax_reset
        Callable[[envs.State, jax.Array], envs.State],  # vmap_brax_step
        Callable[
            [Dict[str, torch.Tensor], torch.Tensor],
            Tuple[Dict[str, torch.Tensor], torch.Tensor],
        ],  # vmap_state_forward
        List[str],  # state_keys
    ],
] = {}


def _evaluate_brax_main(
    env_id: int,
    pop_size: int,
    rotate_key: bool,
    num_episodes: int,
    max_episode_length: int,
    key: torch.Tensor,
    model_state: List[torch.Tensor],
    num_obj: int,
    useless: bool,
    obs_param: torch.Tensor,
    observation_shape: int,
    obs_norm: torch.Tensor,
    obs_buf: torch.Tensor,
    valid_mask: torch.Tensor,
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    # check the pop_size in the inputs
    # Take a parameter and check its size
    actual_pop_size = model_state[0].size(0)
    assert (
        actual_pop_size == pop_size
    ), f"The actual population size must match the pop_size parameter when creating BraxProblem. Expected: {pop_size}, Actual: {actual_pop_size}"
    device = model_state[0].device
    vmap_brax_reset, vmap_brax_step, vmap_state_forward, state_keys = __brax_data__.get(
        env_id
    )
    model_state = {k: v.clone() for k, v in zip(state_keys, model_state)}

    key = to_jax_array(key)
    # For each episode, we need a different random key.
    # For each individual in the population, we need the same set of keys.
    # Loop until environment stops
    if rotate_key:
        key, eval_key = jax.random.split(key)
    else:
        key, eval_key = key, key

    keys = jax.random.split(eval_key, num_episodes)
    keys = jnp.broadcast_to(keys, (pop_size, *keys.shape)).reshape(
        pop_size * num_episodes, -1
    )
    done = jnp.zeros((pop_size * num_episodes,), dtype=bool)
    total_reward = jnp.zeros((pop_size * num_episodes, num_obj))
    counter = 0
    brax_state = vmap_brax_reset(keys)

    while counter < max_episode_length and ~done.all():
        model_state, action = vmap_state_forward(
            model_state,
            from_jax_array(brax_state.obs, device).view(pop_size, num_episodes, -1),
        )
        action = action.view(pop_size * num_episodes, -1)
        clip_val = obs_norm[0]
        std_min = obs_norm[1]
        std_max = obs_norm[2]

        # Perform normalization of the observation space
        if not useless:
            origin_obs = from_jax_array(brax_state.obs, device)
            obs_step = obs_param[0]
            run_var, run_mean = torch.chunk(obs_param[1:], 2)
            run_var = run_var.view((observation_shape,))
            run_mean = run_mean.view((observation_shape,))
            variance = run_var / (obs_step + 1.0)
            variance = clamp(variance, std_min, std_max)
            norm_obs = clamp(
                (origin_obs - run_mean) / torch.sqrt(variance),
                -clip_val,
                clip_val,
            )
            brax_state = brax_state.replace(obs=to_jax_array(norm_obs))
        brax_state = vmap_brax_step(brax_state, to_jax_array(action))

        obs_buf[counter] = from_jax_array(brax_state.obs, device)
        done = jnp.tile(brax_state.done[:, jnp.newaxis], (1, num_obj))
        reward = jnp.nan_to_num(brax_state.reward)
        total_reward += (1 - done) * reward
        jax_vm = to_jax_array(valid_mask[counter].clone())
        valid_mask[counter] = from_jax_array(
            (1 - brax_state.done.ravel()).reshape(jax_vm.shape) * jax_vm, device
        )
        counter += 1

        # Update obs_param
        if not useless:
            obs_step = obs_param[0]
            run_var, run_mean = torch.chunk(obs_param[1:], 2)
            if valid_mask.ndim != obs_buf.ndim:
                valid_mask = valid_mask.view(
                    valid_mask.shape + (1,) * (obs_buf.ndim - valid_mask.ndim)
                )
            new_total_step = obs_step + torch.sum(valid_mask)

            old_mean = (obs_buf - run_mean) * valid_mask
            new_mean = run_mean + torch.sum(old_mean / new_total_step, dim=(0, 1))
            temp_new_mean = (obs_buf - new_mean) * valid_mask
            new_var = run_var + torch.sum(old_mean * temp_new_mean, dim=(0, 1))

            obs_param = torch.concatenate(
                [
                    torch.ones(1, device=new_var.device) * new_total_step,
                    new_var,
                    new_mean,
                ],
                dim=0,
            )

    # Return
    new_key = from_jax_array(key, device)
    total_reward = from_jax_array(total_reward, device)
    total_reward = total_reward.view(pop_size, num_episodes, num_obj)
    model_state = [model_state[k] for k in state_keys]
    return new_key, model_state, total_reward


@torch.library.custom_op("evox::_evaluate_brax", mutates_args=())
def _evaluate_brax(
    env_id: int,
    pop_size: int,
    rotate_key: bool,
    num_episodes: int,
    max_episode_length: int,
    key: torch.Tensor,
    model_state: List[torch.Tensor],
    num_obj: int,
    useless: bool,
    obs_param: torch.Tensor,
    observation_shape: int,
    obs_norm: torch.Tensor,
    obs_buf: torch.Tensor,
    valid_mask: torch.Tensor,
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    return _evaluate_brax_main(
        env_id,
        pop_size,
        rotate_key,
        num_episodes,
        max_episode_length,
        key,
        model_state,
        num_obj,
        useless,
        obs_param,
        observation_shape,
        obs_norm,
        obs_buf,
        valid_mask,
    )


@_evaluate_brax.register_fake
def _fake_evaluate_brax(
    env_id: int,
    pop_size: int,
    rotate_key: bool,
    num_episodes: int,
    max_episode_length: int,
    key: torch.Tensor,
    model_state: List[torch.Tensor],
    num_obj: int,
    useless: bool,
    obs_param: torch.Tensor,
    observation_shape: int,
    obs_norm: torch.Tensor,
    obs_buf: torch.Tensor,
    valid_mask: torch.Tensor,
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    return (
        key.new_empty(key.size()),
        [v.new_empty(v.size()) for v in model_state],
        model_state[0].new_empty(pop_size, num_episodes, num_obj),
    )


@torch.library.custom_op("evox::_evaluate_brax_vmap_main", mutates_args=())
def _evaluate_brax_vmap_main(
    batch_size: int,
    in_dim: List[int],
    env_id: int,
    pop_size: int,
    rotate_key: bool,
    num_episodes: int,
    max_episode_length: int,
    key: torch.Tensor,
    model_state: List[torch.Tensor],
    num_obj: int,
    useless: bool,
    obs_param: torch.Tensor,
    observation_shape: int,
    obs_norm: torch.Tensor,
    obs_buf: torch.Tensor,
    valid_mask: torch.Tensor,
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    # flatten vmap dim and pop dim
    model_state = [
        (v if d is None else v.movedim(d, 0).flatten(0, 1))
        for d, v in zip(in_dim, model_state)
    ]
    key, model_state, reward = _evaluate_brax_main(
        env_id,
        pop_size,
        rotate_key,
        num_episodes,
        max_episode_length,
        key,
        model_state,
        num_obj,
        useless,
        obs_param,
        observation_shape,
        obs_norm,
        obs_buf,
        valid_mask,
    )
    model_state = [
        (v if d is None else v.unflatten(0, (batch_size, -1)))
        for d, v in zip(in_dim, model_state)
    ]
    reward = reward.unflatten(0, (batch_size, -1))
    return key, model_state, reward


@_evaluate_brax.register_vmap
def _evaluate_brax_vmap(
    vmap_info: VmapInfo,
    in_dims: Tuple[int | None | List[int], ...],
    env_id: int,
    pop_size: int,
    rotate_key: bool,
    num_episodes: int,
    max_episode_length: int,
    key: torch.Tensor,
    model_state: List[torch.Tensor],
) -> Tuple[
    Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor],
    Tuple[int | None, List[int], int],
]:
    assert all(d is None for d in in_dims[:-1]), "Cannot vmap over `BraxProblem` itself"
    assert in_dims[-1] is not None, "Cannot vmap none of the dimensions"
    key, model_state, reward = _evaluate_brax_vmap_main(
        vmap_info.batch_size,
        in_dims[-1],
        env_id,
        pop_size,
        rotate_key,
        num_episodes,
        max_episode_length,
        key,
        model_state,
    )
    return (key, model_state, reward), (None, [0] * len(model_state), 0)


@_evaluate_brax_vmap_main.register_fake
def _fake_evaluate_brax_vmap(
    batch_size: int,
    in_dim: List[int],
    env_id: int,
    pop_size: int,
    rotate_key: bool,
    num_episodes: int,
    max_episode_length: int,
    key: torch.Tensor,
    model_state: List[torch.Tensor],
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    return (
        key.new_empty(key.size()),
        [v.new_empty(v.size()).movedim(d, 0) for d, v in zip(in_dim, model_state)],
        model_state[0].new_empty(batch_size, pop_size // batch_size, num_episodes),
    )


class MoRobtrol(Problem):
    """The Brax problem wrapper."""

    def __init__(
        self,
        policy: nn.Module,
        env_name: str,
        max_episode_length: int,
        num_episodes: int,
        seed: int = None,
        pop_size: int | None = None,
        rotate_key: bool = True,
        reduce_fn: Callable[[torch.Tensor, int], torch.Tensor] = torch.mean,
        backend: str | None = None,
        device: torch.device | None = None,
        num_obj: int = 1,
        observation_shape: int = 0,
        obs_norm: torch.Tensor = None,
        useless: bool = True,
    ):
        """Construct a Brax-based problem.
        Firstly, you need to define a policy model.
        Then you need to set the `environment name <https://github.com/google/brax/tree/main/brax/envs>`,
        the maximum episode length, the number of episodes to evaluate for each individual.
        For each individual,
        it will run the policy with the environment for num_episodes times with different seed,
        and use the reduce_fn to reduce the rewards (default to average).
        Different individuals will share the same set of random keys in each iteration.

        :param policy: The policy model whose forward function is :code:`forward(batched_obs) -> action`.
        :param env_name: The environment name.
        :param max_episode_length: The maximum number of time steps of each episode.
        :param num_episodes: The number of episodes to evaluate for each individual.
        :param seed: The seed used to create a PRNGKey for the brax environment. When None, randomly select one. Default to None.
        :param pop_size: The size of the population to be evaluated. If None, we expect the input to have a population size of 1.
        :param rotate_key: Indicates whether to rotate the random key for each iteration (default is True). <br/> If True, the random key will rotate after each iteration, resulting in non-deterministic and potentially noisy fitness evaluations. This means that identical policy weights may yield different fitness values across iterations. <br/> If False, the random key remains the same for all iterations, ensuring consistent fitness evaluations.
        :param reduce_fn: The function to reduce the rewards of multiple episodes. Default to `torch.mean`.
        :param backend: Brax's backend. If None, the default backend of the environment will be used. Default to None.
        :param device: The device to run the computations on. Defaults to the current default device.
        :param num_obj: The number of the objectives. Defaults to 1.
        :param observation_shape: The shape of the observation space. Default to 0.
        :param obs_norm: The observation normalization parameters. The format should be a tensor that represented `[clip_val, std_min, std_max]`. `clip_val` represents the clip interval will be `[-clip_val, clip_val]`, `std_min` represents the minimum standard deviation, `std_max` represents the maximum deviation.

        ## Notice
        The initial key is obtained from `torch.random.get_rng_state()`.

        ## Warning
        This problem does NOT support HPO wrapper (`problems.hpo_wrapper.HPOProblemWrapper`) out-of-box, i.e., the workflow containing this problem CANNOT be vmapped.
        *However*, by setting `pop_size` to the multiplication of inner population size and outer population size, you can still use this problem in a HPO workflow.
        Yet, the `num_repeats` of HPO wrapper *must* be set to 1, please use the parameter `num_episodes` instead.

        ## Examples
        >>> from evox import problems
        >>> problem = problems.neuroevolution.MoRobtrol(
        ...    env_name="swimmer",
        ...    policy=model,
        ...    max_episode_length=1000,
        ...    num_episodes=3,
        ...    pop_size=100,
        ...    num_obj=2,
        ...    observation_shape=8,
        ...    obs_norm=torch.tensor([5.0, 1e-6, 1e6]),
        ...    rotate_key=False,
        ...)
        """
        super().__init__()
        device = torch.get_default_device() if device is None else device
        pop_size = 1 if pop_size is None else pop_size
        # Create Brax environment
        env: envs.Env = (
            envs.get_environment(env_name=env_name)
            if backend is None
            else envs.get_environment(env_name=env_name, backend=backend)
        )
        vmap_env = envs.wrappers.training.VmapWrapper(env)
        # Compile Brax environment
        self.brax_reset = jax.jit(env.reset)
        self.brax_step = jax.jit(env.step)
        self.vmap_brax_reset = jax.jit(vmap_env.reset)
        self.vmap_brax_step = jax.jit(vmap_env.step)
        # JIT stateful model forward
        self.vmap_init_state, self.vmap_state_forward = get_vmap_model_state_forward(
            model=policy,
            pop_size=pop_size,
            in_dims=(0, 0),
            device=device,
        )
        self.state_forward = torch.compile(use_state(policy))
        if seed is None:
            seed = torch.randint(0, 2**31, (1,)).item()
        self.key = from_jax_array(jax.random.PRNGKey(seed), device)
        copied_policy = copy.deepcopy(policy).to(device)
        self.init_state = copied_policy.state_dict()
        for _name, value in self.init_state.items():
            value.requires_grad = False
        # Store to global
        self.state_keys = list(self.init_state.keys())
        global __brax_data__
        __brax_data__[id(self)] = (
            self.vmap_brax_reset,
            self.vmap_brax_step,
            self.vmap_state_forward,
            self.state_keys,
        )
        weakref.finalize(self, __brax_data__.pop, id(self), None)
        # Store variables
        self._id_ = id(self)
        self.reduce_fn = reduce_fn
        self.rotate_key = rotate_key
        self.pop_size = pop_size
        self.num_episodes = num_episodes
        self.max_episode_length = max_episode_length
        self.env_sys = env.sys
        self.device = device

        self.observation_shape = observation_shape
        self.num_obj = num_obj
        if obs_norm is None:
            self.obs_norm = torch.tensor([5.0, 1e-6, 1e6], device=device)
        else:
            self.obs_norm = obs_norm
        self.obs_param = torch.zeros(
            1 + max(self.observation_shape, 1) * 2, device=device
        )
        self.valid_mask = torch.ones(
            (self.max_episode_length, pop_size * self.num_episodes), device=device
        )
        self.obs_buf = torch.zeros(
            (
                self.max_episode_length,
                pop_size * self.num_episodes,
                self.observation_shape,
            ),
            device=device,
        )
        self.useless = useless

    # disable torch.compile for JAX code
    @torch.compiler.disable
    def _evaluate_brax_record(
        self,
        model_state: Dict[str, torch.Tensor],
    ) -> Tuple[Dict[str, torch.Tensor], torch.Tensor, List[Any]]:
        key = to_jax_array(self.key)
        # For each episode, we need a different random key.
        # For each individual in the population, we need the same set of keys.
        # Loop until environment stops
        if self.rotate_key:
            key, eval_key = jax.random.split(key)
        else:
            key, eval_key = key, key

        keys = eval_key
        done = jnp.zeros((), dtype=bool)
        total_reward = jnp.zeros(())
        counter = 0
        brax_state = self.brax_reset(keys)
        trajectory = [brax_state.pipeline_state]

        while counter < self.max_episode_length and ~done.all():
            model_state, action = self.state_forward(
                model_state, from_jax_array(brax_state.obs, self.device)
            )
            brax_state = self.brax_step(brax_state, to_jax_array(action))
            done = brax_state.done * (1 - done)
            total_reward += (1 - done) * brax_state.reward
            counter += 1
            trajectory.append(brax_state.pipeline_state)
        # Return
        self.key = from_jax_array(key, self.device)
        total_reward = from_jax_array(total_reward, self.device)
        return model_state, total_reward, trajectory

    def evaluate(self, pop_params: Dict[str, nn.Parameter]) -> torch.Tensor:
        """Evaluate the final rewards of a population (batch) of model parameters.

        :param pop_params: A dictionary of parameters where each key is a parameter name and each value is a tensor of shape (batch_size, *param_shape) representing the batched parameters of batched models.

        :return: A tensor of shape (pop_size, num_obj) containing the reward of each sample in the population.
        """
        # Merge the given parameters into the initial parameters
        model_state = {**self.vmap_init_state, **pop_params}
        # CANNOT COMPILE: model_state = self.vmap_init_state | pop_params
        model_state = [model_state[k] for k in self.state_keys]
        # Brax environment evaluation
        key, _, rewards = _evaluate_brax(
            env_id=self._id_,
            pop_size=self.pop_size,
            rotate_key=self.rotate_key,
            num_episodes=self.num_episodes,
            max_episode_length=self.max_episode_length,
            key=self.key,
            model_state=model_state,
            num_obj=self.num_obj,
            useless=self.useless,
            obs_param=self.obs_param,
            observation_shape=self.observation_shape,
            obs_norm=self.obs_norm,
            obs_buf=self.obs_buf,
            valid_mask=self.valid_mask,
        )
        self.key = key
        rewards = self.reduce_fn(rewards, dim=1)
        return rewards

    def visualize(
        self,
        weights: Dict[str, nn.Parameter],
        seed: int = 0,
        output_type: str = "HTML",
        *args,
        **kwargs,
    ) -> str | torch.Tensor:
        """Visualize the brax environment with the given policy and weights.

        :param weights: The weights of the policy model. Which is a dictionary of parameters.
        :param output_type: The output type of the visualization, "HTML" or "rgb_array". Default to "HTML".

        :return: The visualization output.
        """
        assert output_type in [
            "HTML",
            "rgb_array",
        ], "output_type must be either HTML or rgb_array"
        model_state = self.init_state | weights
        # Brax environment evaluation
        model_state, _rewards, trajectory = self._evaluate_brax_record(model_state)
        trajectory = [brax_state for brax_state in trajectory]
        if output_type == "HTML":
            return html.render(self.env_sys, trajectory, *args, **kwargs)
        else:
            return image.render_array(self.env_sys, trajectory, **kwargs)



import torch


def hv(objs: torch.Tensor, ref: torch.Tensor, num_sample: int = 100000):
    """
    Monte Carlo Hypervolume Calculation using bounding cube method.

    :param objs: Objective points of shape (n_points, n_objs).
    :param ref: Reference point of shape (n_objs, ).
    :param num_sample: Number of Monte Carlo samples.
    :return: Estimated hypervolume.
    """

    points = torch.abs(objs - ref)
    bound = torch.max(points, dim=0).values
    max_vol = torch.prod(bound)
    samples = torch.rand(num_sample, points.size(1)) * bound
    in_hypercube = torch.any(torch.all(samples.unsqueeze(1) < points.unsqueeze(0), dim=2), dim=1)
    hv = in_hypercube.sum() / num_sample * max_vol
    return hv



import math
from typing import Callable, Optional

import torch
from evox.core import Algorithm, Mutable
from evox.operators.crossover import simulated_binary_half
from evox.operators.mutation import polynomial_mutation
from evox.operators.sampling import uniform_sampling
from evox.utils import clamp, minimum


def pbi(f: torch.Tensor, w: torch.Tensor, z: torch.Tensor):
    norm_w = torch.linalg.norm(w, dim=1)
    f = f - z

    d1 = torch.sum(f * w, dim=1) / norm_w

    d2 = torch.linalg.norm(f - (d1[:, None] * w / norm_w[:, None]), dim=1)
    return d1 + 5 * d2


class MOEAD(Algorithm):
    """
    Implementation of the Original MOEA/D algorithm.

    :references:
        [1] Q. Zhang and H. Li, "MOEA/D: A Multiobjective Evolutionary Algorithm Based on Decomposition,"
            IEEE Transactions on Evolutionary Computation, vol. 11, no. 6, pp. 712-731, 2007. Available:
            https://ieeexplore.ieee.org/document/4358754

    :note: This implementation is based on the original paper and may not be the most efficient implementation.
    """

    def __init__(
        self,
        pop_size: int,
        n_objs: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        selection_op: Optional[Callable] = None,
        mutation_op: Optional[Callable] = None,
        crossover_op: Optional[Callable] = None,
        device: torch.device | None = None,
    ):
        """Initializes the MOEA/D algorithm.

        :param pop_size: The size of the population.
        :param n_objs: The number of objective functions in the optimization problem.
        :param lb: The lower bounds for the decision variables (1D tensor).
        :param ub: The upper bounds for the decision variables (1D tensor).
        :param selection_op: The selection operation for evolutionary strategy (optional).
        :param mutation_op: The mutation operation, defaults to `polynomial_mutation` if not provided (optional).
        :param crossover_op: The crossover operation, defaults to `simulated_binary_half` if not provided (optional).
        :param device: The device on which computations should run (optional). Defaults to PyTorch's default device.
        """
        super().__init__()
        self.pop_size = pop_size
        self.n_objs = n_objs
        if device is None:
            device = torch.get_default_device()
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.shape[0]
        # write to self
        self.lb = lb.to(device=device)
        self.ub = ub.to(device=device)

        self.selection = selection_op
        self.mutation = mutation_op
        self.crossover = crossover_op
        self.device = device

        if self.mutation is None:
            self.mutation = polynomial_mutation
        if self.crossover is None:
            self.crossover = simulated_binary_half

        w, _ = uniform_sampling(self.pop_size, self.n_objs)

        self.pop_size = w.size(0)
        self.n_neighbor = int(math.ceil(self.pop_size / 10))

        length = ub - lb
        population = torch.rand(self.pop_size, self.dim, device=device)
        population = length * population + lb

        neighbors = torch.cdist(w, w)
        self.neighbors = torch.argsort(neighbors, dim=1, stable=True)[:, : self.n_neighbor]
        self.w = w

        self.pop = Mutable(population)
        self.fit = Mutable(torch.empty((self.pop_size, self.n_objs), device=device).fill_(torch.inf))
        self.z = Mutable(torch.zeros((self.n_objs,), device=device))

    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.fit = self.evaluate(self.pop)
        self.z = torch.min(self.fit, dim=0)[0]

    def step(self):
        """Perform a single optimization step of the workflow."""
        for i in range(self.pop_size):
            parents = self.neighbors[i][torch.randperm(self.n_neighbor, device=self.device)]
            crossovered = self.crossover(self.pop[parents[:2]])
            offspring = self.mutation(crossovered, self.lb, self.ub)
            offspring = clamp(offspring, self.lb, self.ub)
            off_fit = self.evaluate(offspring)

            self.z = minimum(self.z, off_fit)

            g_old = pbi(self.fit[parents], self.w[parents], self.z)
            g_new = pbi(off_fit, self.w[parents], self.z)

            self.fit[parents[g_old >= g_new]] = off_fit
            self.pop[parents[g_old >= g_new]] = offspring



__all__ = ["MoRobtrol", "mo_ant", "mo_half_cheetah", "mo_hopper", "mo_hopper_m3", "mo_humanoid",
           "mo_humanoidstandup", "mo_inverted_double_pendulum", "mo_pusher", "mo_reacher",
           "mo_swimmer", "mo_walker2d"]

from evomo.problems.neuroevolution.morobtrol_pro import MoRobtrol

from .morobtrol import (
    mo_ant,
    mo_half_cheetah,
    mo_hopper,
    mo_hopper_m3,
    mo_humanoid,
    mo_humanoidstandup,
    mo_inverted_double_pendulum,
    mo_pusher,
    mo_reacher,
    mo_swimmer,
    mo_walker2d,
)



from typing import Callable, Optional

import torch
from evox.core import Algorithm, Mutable
from evox.operators.crossover import simulated_binary
from evox.operators.mutation import polynomial_mutation
from evox.operators.selection import tournament_selection_multifit
from evox.utils import clamp

from evomo.operators.selection import nd_environmental_selection


class NSGA2(Algorithm):
    """
    A tensorized implementation of the Non-dominated Sorting Genetic Algorithm II (NSGA-II)
    for multi-objective optimization problems.

    :references:
        [1] K. Deb, A. Pratap, S. Agarwal, and T. Meyarivan, "A fast and elitist multiobjective genetic algorithm: NSGA-II,"
            IEEE Transactions on Evolutionary Computation, vol. 6, no. 2, pp. 182-197, 2002.
            Available: https://ieeexplore.ieee.org/document/996017

        [2] Z. Liang, H. Li, N. Yu, K. Sun, and R. Cheng, "Bridging Evolutionary Multiobjective Optimization and
            GPU Acceleration via Tensorization," IEEE Transactions on Evolutionary Computation, 2025. Available:
            https://ieeexplore.ieee.org/document/10944658
    """

    def __init__(
        self,
        pop_size: int,
        n_objs: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        selection_op: Optional[Callable] = None,
        mutation_op: Optional[Callable] = None,
        crossover_op: Optional[Callable] = None,
        device: torch.device | None = None,
    ):
        """Initializes the NSGA-II algorithm.

        :param pop_size: The size of the population.
        :param n_objs: The number of objective functions in the optimization problem.
        :param lb: The lower bounds for the decision variables (1D tensor).
        :param ub: The upper bounds for the decision variables (1D tensor).
        :param selection_op: The selection operation for evolutionary strategy (optional).
        :param mutation_op: The mutation operation, defaults to `polynomial_mutation` if not provided (optional).
        :param crossover_op: The crossover operation, defaults to `simulated_binary` if not provided (optional).
        :param device: The device on which computations should run (optional). Defaults to PyTorch's default device.
        """

        super().__init__()
        self.pop_size = pop_size
        self.n_objs = n_objs
        if device is None:
            device = torch.get_default_device()
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.shape[0]
        # write to self
        self.lb = lb.to(device=device)
        self.ub = ub.to(device=device)

        self.selection = selection_op
        self.mutation = mutation_op
        self.crossover = crossover_op

        if self.selection is None:
            self.selection = tournament_selection_multifit
        if self.mutation is None:
            self.mutation = polynomial_mutation
        if self.crossover is None:
            self.crossover = simulated_binary

        length = ub - lb
        population = torch.rand(self.pop_size, self.dim, device=device)
        population = length * population + lb

        self.pop = Mutable(population)
        self.fit = Mutable(torch.empty((self.pop_size, self.n_objs), device=device).fill_(torch.inf))
        self.rank = Mutable(torch.empty(self.pop_size, device=device).fill_(torch.inf))
        self.dis = Mutable(torch.empty(self.pop_size, device=device).fill_(-torch.inf))

    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.fit = self.evaluate(self.pop)
        _, _, self.rank, self.dis = nd_environmental_selection(self.pop, self.fit, self.pop_size)

    def step(self):
        """Perform the optimization step of the workflow."""
        mating_pool = self.selection(self.pop_size, [-self.dis, self.rank])
        crossovered = self.crossover(self.pop[mating_pool])
        offspring = self.mutation(crossovered, self.lb, self.ub)
        offspring = clamp(offspring, self.lb, self.ub)
        off_fit = self.evaluate(offspring)
        merge_pop = torch.cat([self.pop, offspring], dim=0)
        merge_fit = torch.cat([self.fit, off_fit], dim=0)

        self.pop, self.fit, self.rank, self.dis = nd_environmental_selection(merge_pop, merge_fit, self.pop_size)



__all__ = [
    "mo_ant",
    "mo_half_cheetah",
    "mo_hopper",
    "mo_hopper_m3",
    "mo_humanoid",
    "mo_humanoidstandup",
    "mo_inverted_double_pendulum",
    "mo_pusher",
    "mo_reacher",
    "mo_swimmer",
    "mo_walker2d",
    "register_environment"
]

from brax.envs import register_environment

from evomo.problems.neuroevolution.morobtrol import (
    mo_ant,
    mo_half_cheetah,
    mo_hopper,
    mo_hopper_m3,
    mo_humanoid,
    mo_humanoidstandup,
    mo_inverted_double_pendulum,
    mo_pusher,
    mo_reacher,
    mo_swimmer,
    mo_walker2d,
)

register_environment("mo_halfcheetah", mo_half_cheetah.MoHalfcheetah)
register_environment("mo_hopper_m2", mo_hopper.MoHopper)
register_environment("mo_hopper_m3", mo_hopper_m3.MoHopper)
register_environment("mo_swimmer", mo_swimmer.MoSwimmer)
register_environment("mo_humanoid", mo_humanoid.MoHumanoid)
register_environment(
    "mo_inverted_double_pendulum", mo_inverted_double_pendulum.MoInvertedDoublePendulum
)
register_environment("mo_walker2d", mo_walker2d.MoWalker2d)
register_environment("mo_humanoidstandup", mo_humanoidstandup.MoHumanoidStandup)
register_environment("mo_ant", mo_ant.MoAnt)
register_environment("mo_pusher", mo_pusher.MoPusher)
register_environment("mo_reacher", mo_reacher.MoReacher)



__all__ = ["HypE", "IBEA", "LMOCSO", "MOEAD", "NSGA2", "NSGA3", "RVEA", "RVEAa", "TensorMOEAD"]

from .hype import HypE
from .ibea import IBEA
from .lmocso import LMOCSO
from .moead import MOEAD
from .nsga2 import NSGA2
from .nsga3 import NSGA3
from .rvea import RVEA
from .rveaa import RVEAa
from .tensormoead import TensorMOEAD



import jax.numpy as jnp
from brax import base
from brax.envs.inverted_double_pendulum import InvertedDoublePendulum


class MoInvertedDoublePendulum(InvertedDoublePendulum):
    """
    The multi-objective version of inverted_double_pendulum env.

    :references:
        [1] C. D. Freeman, E. Frey, A. Raichuk, S. Girgin, I. Mordatch, and O. Bachem,
        “Brax - a differentiable physics engine for large scale rigid body simulation,” 2021.
        [Online]. Available: http://github.com/google/brax
    """
    def __init__(self, **kwargs):
        """Initialize the multi-objective inverted_double_pendulum env.

        :param num_obj: The number of the objectives. For this env, it is set to 2.
        """
        super().__init__(**kwargs)
        self.num_obj = 2

    def reset(self, rng):
        state = super().reset(rng)
        mo_reward = jnp.zeros((self.num_obj,))
        return state.replace(reward=mo_reward)

    def step(self, state, action):
        """Run one timestep of the environment's dynamics.

        For more information, please refer to `inverted_double_pendulum <https://github.com/google/brax/tree/main/brax/envs/inverted_double_pendulum.py>` env in brax.
        """
        pipeline_state = self.pipeline_step(state.pipeline_state, action)

        tip = base.Transform.create(pos=jnp.array([0.0, 0.0, 0.6])).do(pipeline_state.x.take(2))
        x, _, y = tip.pos
        dist_penalty = 0.01 * x**2 + (y - 2) ** 2
        v1, v2 = pipeline_state.qd[1:]
        vel_penalty = 1e-3 * v1**2 + 5e-3 * v2**2
        alive_bonus = 10

        obs = self._get_obs(pipeline_state)
        done = jnp.where(y <= 1, jnp.float32(1), jnp.float32(0))

        mo_reward = jnp.array([alive_bonus - dist_penalty, alive_bonus - vel_penalty])
        return state.replace(pipeline_state=pipeline_state, obs=obs, reward=mo_reward, done=done)



from typing import Callable, Optional

import torch
from evox.core import Algorithm, Mutable, vmap
from evox.operators.crossover import simulated_binary
from evox.operators.mutation import polynomial_mutation
from evox.operators.sampling import uniform_sampling
from evox.operators.selection import tournament_selection_multifit
from evox.utils import clamp

from evomo.operators.selection import non_dominate_rank


def _get_table_row_inner(bool_ref_candidate: torch.Tensor, upper_bound: torch.Tensor):
    true_indices = torch.where(
        bool_ref_candidate,
        torch.arange(bool_ref_candidate.size(0), dtype=torch.int32, device=torch.get_default_device()),
        upper_bound,
    )
    true_indices = torch.sort(true_indices, dim=0).values
    return true_indices.to(torch.int32)


vmap_get_table_row = vmap(
    _get_table_row_inner,
    in_dims=(0, None),
)


def _select_from_index_by_min_inner(
    group_id: torch.Tensor,
    group_dist: torch.Tensor,
    idx: torch.Tensor,
):
    min_idx = torch.argmin(torch.where(group_id == idx.unsqueeze(0), group_dist, torch.inf)).to(torch.int32)
    return min_idx


vmap_select_from_index_by_min = vmap(
    _select_from_index_by_min_inner,
    in_dims=(None, None, 0),
)


def _get_extreme_inner(norm_fit: torch.Tensor, w: torch.Tensor):
    return torch.argmin(torch.max(norm_fit / w.unsqueeze(0), dim=1).values)


vmap_get_extreme = vmap(
    _get_extreme_inner,
    in_dims=(None, 0),
)


class NSGA3(Algorithm):
    """
    An implementation of the tensorized NSGA-III for many-objective optimization problems.

    :references:
        [1] K. Deb and H. Jain, "An Evolutionary Many-Objective Optimization Algorithm Using Reference-Point-Based
            Nondominated Sorting Approach, Part I: Solving Problems With Box Constraints," IEEE Transactions on Evolutionary
            Computation, vol. 18, no. 4, pp. 577-601, 2014. Available: https://ieeexplore.ieee.org/document/6600851

        [2] H. Li, Z. Liang, and R. Cheng, "GPU-accelerated Evolutionary Many-objective Optimization Using Tensorized
            NSGA-III," in 2025 IEEE Congress on Evolutionary Computation, 2025.
    """

    def __init__(
        self,
        pop_size: int,
        n_objs: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        selection_op: Optional[Callable] = None,
        mutation_op: Optional[Callable] = None,
        crossover_op: Optional[Callable] = None,
        data_type: Optional[torch.dtype] = None,
        device: torch.device | None = None,
    ):
        """Initializes the NSGA-III algorithm.

        :param pop_size: The size of the population.
        :param n_objs: The number of objective functions in the optimization problem.
        :param lb: The lower bounds for the decision variables (1D tensor).
        :param ub: The upper bounds for the decision variables (1D tensor).
        :param selection_op: The selection operation for evolutionary strategy (optional).
        :param mutation_op: The mutation operation, defaults to `polynomial_mutation` if not provided (optional).
        :param crossover_op: The crossover operation, defaults to `simulated_binary` if not provided (optional).
        :param data_type: The data type for the decision variables (optional). Defaults to torch.float32.
        :param device: The device on which computations should run (optional). Defaults to PyTorch's default device.
        """

        super().__init__()
        self.pop_size = pop_size
        self.n_objs = n_objs
        if device is None:
            device = torch.get_default_device()
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.shape[0]
        # write to self
        self.lb = lb.to(device=device)
        self.ub = ub.to(device=device)

        self.selection = selection_op
        self.mutation = mutation_op
        self.crossover = crossover_op

        if self.selection is None:
            self.selection = tournament_selection_multifit
        if self.mutation is None:
            self.mutation = polynomial_mutation
        if self.crossover is None:
            self.crossover = simulated_binary

        if data_type == torch.bool:
            population = torch.rand(self.pop_size, self.dim, device=device)
            population = population > 0.5
        else:
            length = ub - lb
            population = torch.rand(self.pop_size, self.dim, device=device)
            population = length * population + lb

        self.pop = Mutable(population)
        self.fit = Mutable(torch.full((self.pop_size, self.n_objs), torch.inf, device=device))
        self.rank = Mutable(torch.full((self.pop_size,), torch.inf, device=device))
        self.ref = uniform_sampling(self.pop_size, self.n_objs)[0]

    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.fit = self.evaluate(self.pop)
        self.rank = non_dominate_rank(self.fit)

    def step(self):
        """Perform the optimization step of the workflow."""
        mating_pool = self.selection(self.pop_size, [self.rank])
        crossovered = self.crossover(self.pop[mating_pool])
        offspring = self.mutation(crossovered, self.lb, self.ub)
        offspring = clamp(offspring, self.lb, self.ub)
        off_fit = self.evaluate(offspring)
        merge_pop = torch.cat([self.pop, offspring], dim=0)
        merge_fit = torch.cat([self.fit, off_fit], dim=0)
        shuffled_idx = torch.randperm(merge_pop.shape[0])
        merge_pop = merge_pop[shuffled_idx]
        merge_fit = merge_fit[shuffled_idx]
        rank = non_dominate_rank(merge_fit)
        worst_rank = torch.topk(rank, self.pop_size + 1, largest=False)[0][-1]
        candi_idx = torch.where(rank <= worst_rank)[0]
        merge_pop = merge_pop[candi_idx]
        merge_fit = merge_fit[candi_idx]
        rank = rank[candi_idx]
        device = self.pop.device
        # Normalize
        ideal_point = torch.min(merge_fit, dim=0)[0]
        norm_fit = merge_fit - ideal_point
        weight = torch.eye(self.n_objs, device=device) + 1e-6
        ex_idx = vmap_get_extreme(norm_fit, weight)
        extreme = norm_fit[ex_idx]
        if torch.linalg.matrix_rank(extreme) == self.n_objs:
            hyperplane = torch.linalg.solve(extreme, torch.ones(self.n_objs, device=device))
            intercepts = 1.0 / hyperplane
        else:
            intercepts = torch.max(norm_fit, dim=0).values
        norm_fit = norm_fit / intercepts.unsqueeze(0)
        shuffled_idx = torch.randperm(self.ref.shape[0])
        ref = self.ref[shuffled_idx]
        # Calculate distances by cosine similarity
        distances = self._compute_distances(norm_fit, ref)
        # Associate each solution with its nearest reference point
        group_dist, group_id = torch.min(distances, dim=1)
        # count the number of individuals for each group id
        selected_group_id = group_id[rank < worst_rank]
        rho = torch.bincount(selected_group_id, minlength=ref.shape[0]).to(torch.int32)
        selected_num = torch.sum(rho, dtype=torch.int32)
        candi_group_id = group_id[rank == worst_rank]
        rho_last = torch.bincount(candi_group_id, minlength=ref.shape[0]).to(torch.int32)
        upper_bound = torch.tensor(
            merge_pop.shape[0] + merge_pop.shape[1] + merge_fit.shape[1] + 1, dtype=torch.int32, device=device
        )
        rho = torch.where(rho_last == 0, upper_bound, rho)
        group_id = torch.where(rank == worst_rank, group_id, upper_bound).to(torch.int32)
        row_indices = torch.arange(ref.shape[0], device=device).to(torch.int32)

        # first selection stage
        rho_level = 0
        _selected_ref = rho == rho_level
        selected_ref = torch.where(_selected_ref, row_indices, upper_bound)
        candi_idx = vmap_select_from_index_by_min(group_id, group_dist, selected_ref)
        rank[candi_idx[_selected_ref]] = worst_rank - 1
        rho_last = torch.where(_selected_ref, rho_last - 1, rho_last)
        rho = torch.where(_selected_ref, rho_level + 1, rho)
        rho = torch.where(rho_last == 0, upper_bound, rho)
        selected_num += torch.sum(_selected_ref)

        # second selection stage
        group_id[candi_idx[_selected_ref]] = upper_bound
        bool_ref_candidates = row_indices[:, None] == group_id[None, :]
        ref_candidates = vmap_get_table_row(bool_ref_candidates, upper_bound)
        ref_cand_idx = torch.zeros_like(rho)
        while selected_num < self.pop_size:
            rho_level = torch.min(rho)
            _selected_ref = rho == rho_level
            candi_idx = ref_candidates[row_indices, ref_cand_idx]
            rank[candi_idx[_selected_ref]] = worst_rank - 1
            ref_cand_idx = torch.where(_selected_ref, ref_cand_idx + 1, ref_cand_idx)
            rho_last = torch.where(_selected_ref, rho_last - 1, rho_last)
            rho = torch.where(_selected_ref, rho_level + 1, rho)
            rho = torch.where(rho_last == 0, upper_bound, rho)
            selected_num += torch.sum(_selected_ref)

        # truncate to pop_size
        dif = selected_num - self.pop_size
        candi_idx = torch.where(_selected_ref, candi_idx, upper_bound)
        sorted_index = torch.sort(candi_idx, stable=False)[0]
        rank[sorted_index[:dif]] = worst_rank

        # get final pop and fit
        self.pop = merge_pop[rank < worst_rank]
        self.fit = merge_fit[rank < worst_rank]
        self.rank = rank[rank < worst_rank]

    def _get_extreme(self, norm_fit: torch.Tensor, w: torch.Tensor):
        return torch.argmin(torch.max(norm_fit / w.unsqueeze(0), dim=1).values)

    def _compute_distances(self, fit: torch.Tensor, ref: torch.Tensor):
        # Normalize solutions and reference points to unit vectors
        fit_magnitude = torch.norm(fit, dim=1, keepdim=True).clamp_min(1e-10)
        fit_norm = fit / fit_magnitude
        ref_norm = ref / torch.norm(ref, dim=1, keepdim=True).clamp_min(1e-10)

        # Compute cosine similarity (dot product of normalized vectors)
        cosine_sim = torch.matmul(fit_norm, ref_norm.T)

        # Compute the angular distance component (sqrt(1 - cosine_similarity^2))
        angular_distance = torch.sqrt((1 - cosine_sim**2).clamp_min(1e-10))

        # Compute the final distance by multiplying magnitude with angular distance
        distances = fit_magnitude * angular_distance
        return distances



import torch
import torch.nn.functional as F
from evox.utils import clamp_float, maximum, nanmin


def apd_fn(
    x: torch.Tensor,
    y: torch.Tensor,
    z: torch.Tensor,
    obj: torch.Tensor,
    theta: torch.Tensor,
):
    """
    Compute the APD (Angle-Penalized Distance) based on the given inputs.

    :param x: A tensor representing the indices of the partition.
    :param y: A tensor representing the gamma.
    :param z: A tensor representing the angle.
    :param obj: A tensor of shape (n, m) representing the objectives of the solutions.
    :param theta: A tensor representing the parameter theta used for scaling the reference vector.

    :return: A tensor containing the APD values for each solution.
    """
    selected_z = torch.gather(z, 0, torch.relu(x))
    left = (1 + obj.size(1) * theta * selected_z) / y[None, :]
    norm_obj = torch.linalg.vector_norm(obj, dim=1)
    right = norm_obj[x]
    return left * right


def ref_vec_guided(x: torch.Tensor, f: torch.Tensor, v: torch.Tensor, theta: torch.Tensor):
    """
    Perform the Reference Vector Guided Evolutionary Algorithm (RVEA) selection process.

    This function selects solutions based on the Reference Vector Guided Evolutionary Algorithm.
    It calculates the distances and angles between solutions and reference vectors, and returns
    the next set of solutions to be evolved.

    :param x: A tensor of shape (n, d) representing the current population solutions.
    :param f: A tensor of shape (n, m) representing the objective values for each solution.
    :param v: A tensor of shape (r, m) representing the reference vectors.
    :param theta: A tensor representing the parameter theta used in the APD calculation.

    :return: A tuple containing:
        - next_x: The next selected solutions.
        - next_f: The objective values of the next selected solutions.

    :note:
        The function computes the distances between the solutions and reference vectors,
        and selects the solutions with the minimum APD.
        It currently uses a suboptimal selection implementation, and future improvements
        will optimize the process using a `segment_sort` or `segment_argmin` in CUDA.
    """
    n = f.size(0)
    nv = v.size(0)

    obj = f - nanmin(f, dim=0, keepdim=True)[0]

    obj = maximum(obj, torch.tensor(1e-32, device=f.device))

    cosine = F.cosine_similarity(v.unsqueeze(1), v.unsqueeze(0), dim=-1)

    cosine = torch.where(
        torch.eye(cosine.size(0), dtype=torch.bool, device=f.device),
        0,
        cosine,
    )
    cosine = clamp_float(cosine, 0.0, 1.0)
    gamma = torch.min(torch.acos(cosine), dim=1)[0]

    angle = torch.acos(
        clamp_float(
            F.cosine_similarity(obj.unsqueeze(1), v.unsqueeze(0), dim=-1),
            0.0,
            1.0,
        )
    )

    nan_mask = torch.isnan(obj).any(dim=1)
    associate = torch.argmin(angle, dim=1)
    associate = torch.where(nan_mask, -1, associate)
    associate = associate[:, None]
    partition = torch.arange(0, n, device=f.device)[:, None]
    IndexMatrix = torch.arange(0, nv, device=f.device)[None, :]
    partition = (associate == IndexMatrix) * partition + (associate != IndexMatrix) * -1

    mask = associate != IndexMatrix
    mask_null = mask.sum(dim=0) == n

    apd = apd_fn(partition, gamma, angle, obj, theta)
    apd = torch.where(mask, torch.inf, apd)

    next_ind = torch.argmin(apd, dim=0)
    next_x = torch.where(mask_null.unsqueeze(1), torch.nan, x[next_ind])
    next_f = torch.where(mask_null.unsqueeze(1), torch.nan, f[next_ind])

    return next_x, next_f



__all__ = [
    "ZDT1",
    "ZDT2",
    "ZDT3",
    "ZDT4",
    "ZDT6",
    "MAF1",
    "MAF2",
    "MAF3",
    "MAF4",
    "MAF5",
    "MAF6",
    "MAF7",
    "MAF8",
    "MAF9",
    "MAF10",
    "MAF11",
    "MAF12",
    "MAF13",
    "MAF14",
    "MAF15",
    "LSMOP1",
    "LSMOP2",
    "LSMOP3",
    "LSMOP4",
    "LSMOP5",
    "LSMOP6",
    "LSMOP7",
    "LSMOP8",
    "LSMOP9",
    "DTLZ1",
    "DTLZ2",
    "DTLZ3",
    "DTLZ4",
    "DTLZ5",
    "DTLZ6",
    "DTLZ7",
]

from evomo.problems.numerical.dtlz import DTLZ1, DTLZ2, DTLZ3, DTLZ4, DTLZ5, DTLZ6, DTLZ7
from evomo.problems.numerical.lsmop import LSMOP1, LSMOP2, LSMOP3, LSMOP4, LSMOP5, LSMOP6, LSMOP7, LSMOP8, LSMOP9
from evomo.problems.numerical.maf import (
    MAF1,
    MAF2,
    MAF3,
    MAF4,
    MAF5,
    MAF6,
    MAF7,
    MAF8,
    MAF9,
    MAF10,
    MAF11,
    MAF12,
    MAF13,
    MAF14,
    MAF15,
)
from evomo.problems.numerical.zdt import ZDT1, ZDT2, ZDT3, ZDT4, ZDT6



__all__ = [
    "crowding_distance",
    "nd_environmental_selection",
    "non_dominate_rank",
    "ref_vec_guided",

]


from .non_dominate import crowding_distance, nd_environmental_selection, non_dominate_rank
from .rvea_selection import ref_vec_guided



from functools import partial
from itertools import cycle

import torch
from evox.core import Problem
from evox.operators.sampling import uniform_sampling
from evox.problems.numerical import ackley_func, griewank_func, rastrigin_func, rosenbrock_func, sphere_func


def schwefel_func(x: torch.Tensor) -> torch.Tensor:
    return torch.max(torch.abs(x), dim=-1).values


class LSMOP(Problem):
    """
    The LSMOP test suite for large-scale multiobjective and many-objective optimization.

    :references:
        [1] R. Cheng, Y. Jin, M. Olhofer, and B. Sendhoff, "Test Problems for Large-Scale
            Multiobjective and Many-Objective Optimization," IEEE Transactions on Cybernetics,
            vol. 47, no. 12, pp. 4108-4121, 2017. Available: https://ieeexplore.ieee.org/abstract/document/7553457
    """


    def __init__(self, d=None, m=None, ref_num=1000):
        """
        :param d: number of decision variables
        :param m: number of objectives
        :param ref_num: number of reference points, defaults to 1000
        """
        super().__init__()
        self.nk = 5
        self.m = 3 if m is None else m
        self.d = self.m * 100 if d is None else d
        self.ref_num = ref_num
        self._calc_pf()

        # Calculate the number of subgroup and their length
        c = [3.8 * 0.1 * (1 - 0.1)]
        for i in range(1, self.m):
            c.append(3.8 * c[-1] * (1 - c[-1]))
        c = torch.tensor(c, dtype=torch.float32)
        self.sublen = torch.floor(c / torch.sum(c) * (self.d - self.m + 1) / self.nk)
        self.len = torch.cat([torch.tensor([0]), torch.cumsum(self.sublen * self.nk, dim=0)], dim=0)
        self.sublen = tuple(map(int, self.sublen))
        self.len = tuple(map(int, self.len))

        self.sphere = sphere_func
        self.griewank = griewank_func
        self.rosenbrock = rosenbrock_func
        self.ackley = partial(ackley_func, a=20, b=0.2, c=2 * torch.pi)
        self.rastrigin = rastrigin_func
        self.schwefel = schwefel_func

    def evaluate(self, X: torch.Tensor):
        X_ = self._calc_X_(X)
        g = self._calc_g(X_)
        return self._calc_f(X_, g)

    def pf(self):
        return self._pf_value

    def _calc_pf(self):
        f = uniform_sampling(self.ref_num * self.m, self.m)[0] / 2
        self._pf_value = f

    def _calc_X_(self, X: torch.Tensor):
        m = self.m
        n, d = X.size()
        X_ = X.clone()
        X_[:, m - 1 :] = (1 + (torch.arange(m, d + 1, device=X_.device) / d).unsqueeze(0).repeat(n, 1)) * X_[:, m - 1 :] - X_[
            :, :1
        ] * 10
        return X_

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.sphere], X)

    def _calc_f(self, X_: torch.Tensor, g):
        m = self.m
        n, d = X_.size()
        ones_col = torch.ones(n, 1, device=X_.device)
        cumprod_part = torch.cumprod(torch.cat([ones_col, X_[:, : m - 1]], dim=1), dim=1)
        f = (1 + g) * torch.flip(cumprod_part, [1]) * torch.cat([ones_col, 1 - torch.flip(X_[:, : m - 1], [1])], dim=1)
        return f

    def _inner_calc_g(self, inner_funcs, x):
        g_list = []

        for len_, sublen, func in zip(self.len, self.sublen, cycle(inner_funcs)):
            g_sum = []
            for j in range(self.nk):
                start = len_ + self.m - 1 + j * sublen
                end = start + sublen
                X_slice = x[:, start:end]
                g_sum.append(func(x=X_slice))
            g_list.append(torch.stack(g_sum, dim=1).sum(dim=1))

        g = torch.stack(g_list, dim=1)
        sublen_tensor = torch.tensor(self.sublen, device=x.device).float()
        g = g / sublen_tensor / self.nk
        return g


class LSMOP1(LSMOP):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)


class LSMOP2(LSMOP):
    def __init__(self, d=None, m=3, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.griewank, self.schwefel], X)


class LSMOP3(LSMOP):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.rastrigin, self.rosenbrock], X)


class LSMOP4(LSMOP):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.ackley, self.griewank], X)


class LSMOP5(LSMOP):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_X_(self, X):
        m = self.m
        n, d = X.size()
        X_ = X.clone()
        X_[:, m - 1 :] = (
            1 + torch.cos(torch.arange(m, d + 1, device=X_.device) / d * torch.pi / 2).unsqueeze(0).repeat(n, 1)
        ) * X_[:, m - 1 :] - X_[:, :1] * 10
        return X_

    def _calc_f(self, X_: torch.Tensor, g):
        m = self.m
        n, d = X_.size()
        ones_col = torch.ones(n, 1, device=X_.device)
        cumprod_part = torch.cumprod(torch.cat([ones_col, torch.cos(X_[:, : m - 1] * torch.pi / 2)], dim=1), dim=1)
        last_part = torch.sin(torch.flip(X_[:, : m - 1], [1]) * torch.pi / 2)
        f = (
            (1 + g + torch.cat([g[:, 1:], torch.zeros(n, 1, device=X_.device)], dim=1))
            * torch.flip(cumprod_part, [1])
            * torch.cat([ones_col, last_part], dim=1)
        )
        return f

    def _calc_pf(self):
        f = uniform_sampling(self.ref_num * self.m, self.m)[0] / 2
        self._pf_value = f / torch.sqrt(torch.sum(f**2, dim=1, keepdims=True))


class LSMOP6(LSMOP5):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.rosenbrock, self.schwefel], X)


class LSMOP7(LSMOP5):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.ackley, self.rosenbrock], X)


class LSMOP8(LSMOP5):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        return self._inner_calc_g([self.griewank, self.sphere], X)


class LSMOP9(LSMOP):
    def __init__(self, d=None, m=None, ref_num=1000):
        super().__init__(d, m, ref_num)

    def _calc_g(self, X: torch.Tensor):
        g = self._inner_calc_g([self.sphere, self.ackley], X)
        return 1 + torch.sum(g, dim=1, keepdims=True)

    def _calc_f(self, X_: torch.Tensor, g):
        m = self.m
        n, d = X_.size()
        f = torch.zeros(n, m, device=X_.device)
        f[:, : m - 1] = X_[:, : m - 1]
        f[:, m - 1 :] = (1 + g) * (
            m
            - torch.sum(
                f[:, : m - 1] / (1 + g) * (1 + torch.sin(3 * torch.pi * f[:, : m - 1])),
                dim=1,
                keepdims=True,
            )
        )
        return f

    def _calc_pf(self):
        interval = [0, 0.251412, 0.631627, 0.859401]
        median = (interval[1] - interval[0]) / (interval[3] - interval[2] + interval[1] - interval[0])
        N = self.ref_num * self.m
        X = self._grid(N, self.m - 1)
        X = torch.where(
            X <= median,
            X * (interval[1] - interval[0]) / median + interval[0],
            (X - median) * (interval[3] - interval[2]) / (1 - median) + interval[2],
        )
        p = torch.cat(
            [
                X,
                2 * (self.m - torch.sum(X / 2 * (1 + torch.sin(3 * torch.pi * X)), dim=1, keepdims=True)),
            ],
            dim=1,
        )
        self._pf_value = p

    def _grid(self, N, M):
        gap = torch.linspace(0, 1, int(torch.ceil(torch.tensor(N ** (1 / M)))))
        c = torch.meshgrid(*([gap] * M), indexing="ij")
        w = torch.stack(c, dim=1).reshape(-1, M)
        return w



from typing import Callable, Optional

import torch
from evox.core import Algorithm, Mutable, Parameter
from evox.operators.mutation import polynomial_mutation
from evox.operators.sampling import uniform_sampling
from evox.operators.selection import ref_vec_guided
from evox.utils import clamp, randint


class LMOCSO(Algorithm):
    """
        The tensorized version of LMOCSO algorithm.

    :reference:
        [1] Ye Tian, Xiutao Zheng, Xingyi Zhang and Yaochu Jin, "Efficient Large-Scale Multiobjective Optimization Based
            on a Competitive Swarm," in IEEE Transactions on Cybernetics , 2020, pp. 3696 - 3708. Available:
            https://ieeexplore.ieee.org/document/8681243

    """

    def __init__(
        self,
        n_objs: int,
        pop_size: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        alpha: float = 2.0,
        max_gen: int = 100,
        mutation_op: Optional[Callable] = None,
        selection_op: Optional[Callable] = None,
        device: torch.device | None = None,
    ):
        """
        Initializes the LMOCSO algorithm.

        :param n_objs: The number of objective functions in the optimization problem.
        :param pop_size: The size of the population.
        :param lb: The lower bounds for the decision variables (1D tensor).
        :param ub: The upper bounds for the decision variables (1D tensor).
        :param alpha: The parameter controlling the rate of change of penalty in RVEA selection. Defaults to 2.0.
        :param max_gen: The maximum number of generations for the optimization process. Defaults to 100.
        :param mutation_op: The mutation operation, defaults to `polynomial_mutation` if not provided (optional).
        :param selection_op: The selection operation, defaults to `ref_vec_guided` (Reference Vector Guided Selection) if not provided (optional).
        :param device: The device on which computations should run (optional). Defaults to PyTorch's default device.
        """
        super().__init__()

        if device is None:
            device = torch.get_default_device()
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.shape[0]
        # write to self
        self.n_objs = n_objs
        self.pop_size = pop_size
        self.lb = lb.to(device=device)
        self.ub = ub.to(device=device)
        self.alpha = Parameter(alpha)
        self.max_gen = max_gen

        self.selection = selection_op
        self.mutation = mutation_op

        if self.selection is None:
            self.selection = ref_vec_guided
        if self.mutation is None:
            self.mutation = polynomial_mutation

        reference_vector, _ = uniform_sampling(n=self.pop_size, m=self.n_objs)
        reference_vector = reference_vector.to(device=device)
        self.pop_size = reference_vector.size(0)

        population = torch.rand(self.pop_size, self.dim, device=device)
        population = population * (self.ub - self.lb) + self.lb
        self.pop = Mutable(population)
        self.velocity = Mutable(torch.zeros(self.pop_size // 2 * 2, self.dim, device=device))

        self.reference_vector = Mutable(reference_vector)
        self.fit = Mutable(torch.full((self.pop_size, self.n_objs), torch.inf, device=device))
        self.gen = Mutable(torch.tensor(0, dtype=int, device=device))

    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.fit = self.evaluate(self.pop)

    def step(self):
        """Perform the optimization step of the workflow."""

        valid_mask = ~torch.isnan(self.pop).all(axis=1)
        num_valid = torch.sum(valid_mask, dtype=torch.int32)
        mating_pool = randint(0, num_valid, (self.pop_size,), device=self.pop.device)

        sorted_indices = torch.where(
            valid_mask,
            torch.arange(self.pop_size, device=self.pop.device),
            torch.iinfo(torch.int32).max,
        )
        sorted_indices = torch.argsort(sorted_indices, stable=True)
        selected_pop = self.pop[sorted_indices[mating_pool]]
        selected_fit = self.fit[sorted_indices[mating_pool]]

        sde_fitness = self.cal_fitness(selected_fit)

        randperm = torch.randperm(self.pop_size // 2 * 2, device=self.pop.device).reshape(2, -1)

        mask = sde_fitness[randperm[0, :]] > sde_fitness[randperm[1, :]]
        winner = torch.where(mask, randperm[0, :], randperm[1, :])
        loser = torch.where(mask, randperm[1, :], randperm[0, :])

        r0 = torch.rand(self.pop_size // 2, self.dim, device=self.pop.device)
        r1 = torch.rand(self.pop_size // 2, self.dim, device=self.pop.device)

        off_velocity = r0 * self.velocity[loser] + r1 * (selected_pop[winner] - selected_pop[loser])
        new_loser_population = clamp(
            selected_pop[loser] + off_velocity + r0 * (off_velocity - self.velocity[loser]),
            self.lb,
            self.ub,
        )

        new_population = selected_pop.clone()
        new_population[loser] = new_loser_population
        new_velocity = self.velocity.clone()
        new_velocity[loser] = off_velocity
        self.velocity = new_velocity

        next_generation = self.mutation(new_population, self.lb, self.ub)
        next_generation_fitness = self.evaluate(next_generation)

        self.gen = self.gen + 1

        merged_pop = torch.cat([self.pop, next_generation], dim=0)
        merged_fitness = torch.cat([self.fit, next_generation_fitness], dim=0)
        # RVEA Selection
        survivor, survivor_fitness = self.selection(merged_pop, merged_fitness, self.reference_vector,
                                                    (self.gen / self.max_gen) ** self.alpha)

        self.pop = survivor
        self.fit = survivor_fitness

    def cal_fitness(self, obj):
        """
        Calculate the fitness by shift-based density
        """
        n = obj.shape[0]
        f_max, _ = torch.max(obj, dim=0, keepdim=True)
        f_min, _ = torch.min(obj, dim=0, keepdim=True)
        f = (obj - f_min) / (f_max - f_min + 1e-10)
        s_obj = torch.maximum(
            f.unsqueeze(1),
            f.unsqueeze(0)
        )
        dis = torch.norm(
            f.unsqueeze(1) - s_obj,
            p=2, dim=2
        )
        dis = dis + torch.diag(torch.full((n,), float('inf'), device=obj.device))
        fitness, _ = torch.min(dis, dim=1)

        return fitness




from typing import Callable, Optional

import torch
import torch.nn.functional as F
from evox.core import Algorithm, Mutable, Parameter
from evox.operators.crossover import simulated_binary
from evox.operators.mutation import polynomial_mutation
from evox.operators.sampling import uniform_sampling
from evox.utils import clamp, nanmax, nanmin, randint

from evomo.operators.selection import non_dominate_rank, ref_vec_guided


class RVEAa(Algorithm):
    """
    An implementation of the Reference Vector Guided Evolutionary Algorithm embedded with the reference vector
    regeneration strategy (RVEAa) for multi-objective optimization problems.

    This class is designed to solve multi-objective optimization problems using a reference vector guided evolutionary algorithm.

    :references:
        [1] R. Cheng, Y. Jin, M. Olhofer, and B. Sendhoff, "A reference vector guided evolutionary algorithm
            for many-objective optimization," IEEE Transactions on Evolutionary Computation, vol. 20, no. 5,
            pp. 773-791, 2016. Available: https://ieeexplore.ieee.org/document/7386636
    """

    def __init__(
        self,
        pop_size: int,
        n_objs: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        alpha: float = 2.0,
        fr: float = 0.1,
        max_gen: int = 100,
        selection_op: Optional[Callable] = None,
        mutation_op: Optional[Callable] = None,
        crossover_op: Optional[Callable] = None,
        device: torch.device | None = None,
    ):
        """Initialize the RVEAa algorithm with the given parameters.

        :param pop_size: The size of the population.
        :param n_objs: The number of objective functions in the optimization problem.
        :param lb: The lower bounds for the decision variables.
        :param ub: The upper bounds for the decision variables.
        :param alpha: A parameter for controlling the rate of change of penalty. Defaults to 2.
        :param fr: The frequency of reference vector adaptation. Defaults to 0.1.
        :param max_gen: The maximum number of generations. Defaults to 100.
        :param selection_op: The selection operation for evolutionary strategy (optional).
        :param mutation_op: The mutation operation (optional).
        :param crossover_op: The crossover operation (optional).
        :param device: The device on which computations should run (optional).
        """
        super().__init__()
        self.pop_size = pop_size
        self.n_objs = n_objs
        device = torch.get_default_device() if device is None else device
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.size(0)
        # write to self
        self.lb = lb.unsqueeze(0).to(device=device)
        self.ub = ub.unsqueeze(0).to(device=device)

        self.alpha = Parameter(alpha)
        self.fr = Parameter(fr)
        self.max_gen = Parameter(max_gen)

        self.rv_adapt_every = Mutable(torch.max(torch.round(1 / self.fr), torch.tensor(1.0)))

        self.selection = selection_op
        self.mutation = mutation_op
        self.crossover = crossover_op

        if self.selection is None:
            self.selection = ref_vec_guided
        if self.mutation is None:
            self.mutation = polynomial_mutation
        if self.crossover is None:
            self.crossover = simulated_binary
        sampling, _ = uniform_sampling(self.pop_size, self.n_objs)

        v = sampling.to(device=device)

        v0 = v.clone()
        self.pop_size = v.size(0)
        length = self.ub - self.lb
        population = torch.rand(self.pop_size, self.dim, device=device)
        population = length * population + self.lb
        v1 = torch.rand(self.pop_size, self.n_objs, device=device)
        v = torch.cat([v, v1], dim=0)

        self.pop = Mutable(population)
        self.fit = Mutable(torch.empty((self.pop_size, self.n_objs), device=device).fill_(torch.inf))
        self.reference_vector = Mutable(v)
        self.init_v = v0
        self.gen = Mutable(torch.tensor(0, dtype=int, device=device))

    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.rv_adapt_every = torch.max(torch.round(1 / self.fr), torch.tensor(1.0))
        self.fit = self.evaluate(self.pop)

    def _rv_adaptation(self, pop_obj: torch.Tensor):
        max_vals = nanmax(pop_obj, dim=0)[0]
        min_vals = nanmin(pop_obj, dim=0)[0]
        return self.init_v * (max_vals - min_vals)

    def _no_rv_adaptation(self, pop_obj: torch.Tensor):
        return self.reference_vector[: self.pop_size].clone()

    def _mating_pool(self):
        valid_mask = ~torch.isnan(self.pop).all(dim=1)
        num_valid = torch.sum(valid_mask, dtype=torch.int32)
        mating_pool = randint(0, num_valid, (self.pop_size,), device=self.pop.device)
        sorted_indices = torch.where(valid_mask, torch.arange(self.pop.size(0), device=self.pop.device), torch.iinfo(torch.int32).max)
        sorted_indices = torch.argsort(sorted_indices, stable=True)
        pop = self.pop[sorted_indices[mating_pool]]
        return pop

    def _rv_regeneration(self, pop_obj: torch.Tensor, v: torch.Tensor):
        pop_obj = pop_obj - nanmin(pop_obj, dim=0).values
        cosine = F.cosine_similarity(pop_obj.unsqueeze(1), v.unsqueeze(0), dim=-1)

        mask = torch.isnan(cosine)
        input_tensor = torch.where(mask, -torch.inf, cosine)
        associate = input_tensor.max(dim=1, keepdim=False).indices
        associate = torch.where(input_tensor[:, 0] == -torch.inf, -1, associate)

        invalid = torch.sum((associate.unsqueeze(1) == torch.arange(v.size(0), device=pop_obj.device)), dim=0)
        rand = torch.rand((v.size(0), v.size(1)), device=pop_obj.device) * nanmax(pop_obj, dim=0).values
        new_v = torch.where((invalid == 0).unsqueeze(1), rand, v)

        return new_v

    def _batch_truncation(self, pop: torch.Tensor, obj: torch.Tensor):
        n = pop.size(0) // 2
        cosine = F.cosine_similarity(obj.unsqueeze(1), obj.unsqueeze(0), dim=-1)
        not_all_nan_rows = ~torch.isnan(cosine).all(dim=1)
        mask = torch.eye(cosine.size(0), dtype=torch.bool, device=pop.device) & not_all_nan_rows.unsqueeze(1)
        cosine = torch.where(mask, 0, cosine)

        sorted_values, _ = torch.sort(-cosine, dim=1)
        sorted_values = torch.where(torch.isnan(sorted_values[:, 0]), -torch.inf, sorted_values[:, 0])
        rank = torch.argsort(sorted_values)

        mask = torch.ones(rank.size(0), dtype=torch.bool, device=pop.device)
        mask = torch.where(torch.arange(rank.size(0), device=pop.device) < n, torch.tensor(0, dtype=torch.bool, device=pop.device), mask)
        mask = mask.unsqueeze(1)

        new_pop = torch.where(mask, pop, torch.nan)
        new_obj = torch.where(mask, obj, torch.nan)

        return new_pop, new_obj

    def _no_batch_truncation(self, pop: torch.Tensor, obj: torch.Tensor):
        return pop.clone(), obj.clone()

    def _update_pop_and_rv(self, survivor: torch.Tensor, survivor_fit: torch.Tensor):
        v_regen = self._rv_regeneration(survivor_fit, self.reference_vector[self.pop_size :])
        if torch.compiler.is_compiling():
            v_adapt = torch.cond(
                self.gen % self.rv_adapt_every == 0, self._rv_adaptation, self._no_rv_adaptation, (survivor_fit,)
            )
            self.pop, self.fit = torch.cond(self.gen == self.max_gen, self._batch_truncation, self._no_batch_truncation, (survivor, survivor_fit))
        else:
            if self.gen % self.rv_adapt_every == 0:
                v_adapt = self._rv_adaptation(survivor_fit)
            else:
                v_adapt = self._no_rv_adaptation(survivor_fit)
            if self.gen == self.max_gen:
                self.pop, self.fit = self._batch_truncation(survivor, survivor_fit)
            else:
                self.pop, self.fit = self._no_batch_truncation(survivor, survivor_fit)
        self.reference_vector = torch.cat([v_adapt, v_regen], dim=0)

    def step(self):
        """Perform a single optimization step."""

        self.gen = self.gen + 1
        pop = self._mating_pool()
        crossovered = self.crossover(pop)
        offspring = self.mutation(crossovered, self.lb, self.ub)
        offspring = clamp(offspring, self.lb, self.ub)
        off_fit = self.evaluate(offspring)
        merge_pop = torch.cat([self.pop, offspring], dim=0)
        merge_fit = torch.cat([self.fit, off_fit], dim=0)

        rank = non_dominate_rank(merge_fit)
        merge_fit = torch.where(rank.unsqueeze(1) == 0, merge_fit, torch.nan)
        merge_pop = torch.where(rank.unsqueeze(1) == 0, merge_pop, torch.nan)

        survivor, survivor_fit = self.selection(
            merge_pop,
            merge_fit,
            self.reference_vector,
            (self.gen / self.max_gen) ** self.alpha,
        )

        self._update_pop_and_rv(survivor, survivor_fit)



import torch
from evox.utils import lexsort, register_vmap_op


def dominate_relation(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Return the domination relation matrix A, where A_{ij} is True if x_i dominates y_j.

    :param x: An array with shape (n1, m) where n1 is the population size and m is the number of objectives.
    :param y: An array with shape (n2, m) where n2 is the population size and m is the number of objectives.

    :returns: The domination relation matrix of x and y.
    """
    # Expand the dimensions of x and y so that we can perform element-wise comparisons
    # Add new dimensions to x and y to prepare them for broadcasting
    x_expanded = x.unsqueeze(1)  # Shape (n1, 1, m)
    y_expanded = y.unsqueeze(0)  # Shape (1, n2, m)

    # Broadcasted comparison: each pair (x_i, y_j)
    less_than_equal = x_expanded <= y_expanded  # Shape (n1, n2, m)
    strictly_less_than = x_expanded < y_expanded  # Shape (n1, n2, m)

    # Check the domination condition: x_i dominates y_j
    domination_matrix = less_than_equal.all(dim=2) & strictly_less_than.any(dim=2)

    return domination_matrix


def update_dc_and_rank(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    pareto_front: torch.BoolTensor,
    rank: torch.Tensor,
    current_rank: int,
):
    """
    Update the dominate count and ranks for the current Pareto front.

    :param dominate_relation_matrix: The domination relation matrix between individuals.
    :param dominate_count: The count of how many individuals dominate each individual.
    :param pareto_front: A tensor indicating which individuals are in the current Pareto front.
    :param rank: A tensor storing the rank of each individual.
    :param current_rank: The current Pareto front rank.

    :returns:
        - **rank**: Updated rank tensor.
        - **dominate_count**: Updated dominate count tensor.
    """

    # Update the rank for individuals in the Pareto front
    rank = torch.where(pareto_front, current_rank, rank)
    # Calculate how many individuals in the Pareto front dominate others
    count_desc = torch.sum(pareto_front.unsqueeze(-1) * dominate_relation_matrix, dim=-2)

    # Update dominate_count (remove those in the current Pareto front)
    dominate_count = dominate_count - count_desc
    dominate_count = dominate_count - pareto_front.int()

    return rank, dominate_count


def _igr_fake(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    rank: torch.Tensor,
    pareto_front: torch.Tensor,
    compiling: bool,
) -> torch.Tensor:
    return rank.new_empty(dominate_count.size())


def _igr_fake_vmap(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    rank: torch.Tensor,
    pareto_front: torch.Tensor,
    compiling: bool,
) -> torch.Tensor:
    return rank.new_empty(dominate_count.size())


def _vmap_iterative_get_ranks_compile(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    rank: torch.Tensor,
    pareto_front: torch.Tensor,
) -> torch.Tensor:
    def cond_fn(r, cr, dc, pf):
        return pf.any()

    def body_fn(r, cr, dc, pf):
        r, dc = update_dc_and_rank(dominate_relation_matrix, dc, pf, r, cr)
        cr = cr + 1
        new_pareto_front = dc == 0
        pf = torch.where(pf.any(dim=-1, keepdim=True), new_pareto_front, pf)
        return r, cr, dc, pf

    rank = rank.expand_as(dominate_count).contiguous() # contiguous to unify carry stride
    rank, *_ = torch.while_loop(
        cond_fn, body_fn, (rank, torch.tensor(0, device=rank.device), dominate_count, pareto_front)
    )
    return rank


# evox.core.compile is not necessary since no indexing here
_vmap_iterative_get_ranks_compile = torch.compile(_vmap_iterative_get_ranks_compile, fullgraph=True)


def _vmap_iterative_get_ranks(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    rank: torch.Tensor,
    pareto_front: torch.Tensor,
    compiling: bool,
) -> torch.Tensor:
    current_rank = 0
    if compiling:
        rank = _vmap_iterative_get_ranks_compile(dominate_relation_matrix, dominate_count, rank, pareto_front)
    else:
        while pareto_front.any():
            rank, dominate_count = update_dc_and_rank(
                dominate_relation_matrix, dominate_count, pareto_front, rank, current_rank
            )
            current_rank += 1
            new_pareto_front = dominate_count == 0
            pareto_front = torch.where(pareto_front.any(dim=-1, keepdim=True), new_pareto_front, pareto_front)
    return rank


def _iterative_get_ranks_compile(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    rank: torch.Tensor,
    pareto_front: torch.Tensor,
) -> torch.Tensor:
    def cond_fn(r, cr, dc, pf):
        return pf.any()

    def body_fn(r, cr, dc, pf):
        r, dc = update_dc_and_rank(dominate_relation_matrix, dc, pf, r, cr)
        cr = cr + 1
        pf = dc == 0
        return r, cr, dc, pf

    rank, *_ = torch.while_loop(
        cond_fn, body_fn, (rank, torch.tensor(0, device=rank.device), dominate_count, pareto_front)
    )
    return rank


# evox.core.compile is not necessary since no indexing here
_iterative_get_ranks_compile = torch.compile(_iterative_get_ranks_compile, fullgraph=True)


@register_vmap_op(
    fake_fn=_igr_fake, vmap_fn=_vmap_iterative_get_ranks, fake_vmap_fn=_igr_fake_vmap, max_vmap_level=2
)
def _iterative_get_ranks(
    dominate_relation_matrix: torch.Tensor,
    dominate_count: torch.Tensor,
    rank: torch.Tensor,
    pareto_front: torch.Tensor,
    compiling: bool,
) -> torch.Tensor:
    if compiling:
        rank = _iterative_get_ranks_compile(dominate_relation_matrix, dominate_count, rank, pareto_front)
    else:
        current_rank = 0
        while pareto_front.any():
            rank, dominate_count = update_dc_and_rank(
                dominate_relation_matrix, dominate_count, pareto_front, rank, current_rank
            )
            current_rank += 1
            pareto_front = dominate_count == 0
    return rank


def non_dominate_rank(x: torch.Tensor) -> torch.Tensor:
    """
    Compute the non-domination rank for a set of solutions in multi-objective optimization.

    The non-domination rank is a measure of the Pareto optimality of each solution.

    :param f: A 2D tensor where each row represents a solution, and each column represents an objective.

    :returns:
        A 1D tensor containing the non-domination rank for each solution.
    """

    n = x.size(0)
    # Domination relation matrix (n x n)
    dominate_relation_matrix = dominate_relation(x, x)
    # Count how many times each individual is dominated
    dominate_count = dominate_relation_matrix.sum(dim=0)
    # Initialize rank array
    rank = torch.zeros(n, dtype=torch.int32, device=x.device)
    # Identify individuals in the first Pareto front (those that are not dominated)
    pareto_front = dominate_count == 0
    # Iteratively identify Pareto fronts
    rank = _iterative_get_ranks(
        dominate_relation_matrix, dominate_count, rank, pareto_front, torch.compiler.is_compiling()
    )
    return rank


def crowding_distance(costs: torch.Tensor, mask: torch.Tensor):
    """
    Compute the crowding distance for a set of solutions in multi-objective optimization.

    The crowding distance is a measure of the diversity of solutions within a Pareto front.

    :param costs: A 2D tensor where each row represents a solution, and each column represents an objective.
    :param mask: A 1D boolean tensor indicating which solutions should be considered.

    :returns:
        A 1D tensor containing the crowding distance for each solution.
    """
    total_len = costs.size(0)
    if mask is None:
        num_valid_elem = total_len
        mask = torch.ones(total_len, dtype=torch.bool)
    else:
        num_valid_elem = mask.sum()

    inverted_mask = ~mask

    inverted_mask = inverted_mask.unsqueeze(1).expand(-1, costs.size(1)).to(costs.dtype)

    rank = lexsort([costs, inverted_mask], dim=0)
    costs = torch.gather(costs, dim=0, index=rank)
    distance_range = costs[num_valid_elem - 1] - costs[0]
    distance = torch.empty(costs.size(), device=costs.device)
    distance = distance.scatter(0, rank[1:-1], (costs[2:] - costs[:-2]) / distance_range)
    distance[rank[0], :] = torch.inf
    distance[rank[num_valid_elem - 1], :] = torch.inf
    crowding_distances = torch.where(mask.unsqueeze(1), distance, -torch.inf)
    crowding_distances = torch.sum(crowding_distances, dim=1)

    return crowding_distances


def nd_environmental_selection(x: torch.Tensor, f: torch.Tensor, topk: int):
    """
    Perform environmental selection based on non-domination rank and crowding distance.

    :param x: A 2D tensor where each row represents a solution, and each column represents a decision variable.
    :param f: A 2D tensor where each row represents a solution, and each column represents an objective.
    :param topk: The number of solutions to select.

    :returns:
        A tuple of four tensors:
        - **x**: The selected solutions.
        - **f**: The corresponding objective values.
        - **rank**: The non-domination rank of the selected solutions.
        - **crowding_dis**: The crowding distance of the selected solutions.
    """
    rank = non_dominate_rank(f)
    worst_rank = torch.topk(rank, topk, largest=False)[0][-1]
    mask = rank == worst_rank
    crowding_dis = crowding_distance(f, mask)
    combined_order = lexsort([-crowding_dis, rank])[:topk]
    return x[combined_order], f[combined_order], rank[combined_order], crowding_dis[combined_order]



import jax
import jax.numpy as jnp
from brax.envs.base import State
from brax.envs.humanoidstandup import HumanoidStandup


class MoHumanoidStandup(HumanoidStandup):
    """
    The multi-objective version of humanoidstandup.

    :references:
        [1] C. D. Freeman, E. Frey, A. Raichuk, S. Girgin, I. Mordatch, and O. Bachem,
        “Brax - a differentiable physics engine for large scale rigid body simulation,” 2021.
        [Online]. Available: http://github.com/google/brax
    """
    def __init__(self, **kwargs):
        """Initialize the multi-objective humanoidstandup env.

        :param num_obj: The number of the objectives. For this env, it is set to 2.
        """
        super().__init__(**kwargs)
        self.num_obj = 2

    def reset(self, rng: jax.Array):
        state = super().reset(rng)
        reward = jnp.zeros((self.num_obj,))
        return state.replace(reward=reward)

    def step(self, state: State, action: jax.Array):
        """Run one timestep of the environment's dynamics.

        :param quad_energy_cost: The energy consumed by the control force of the humanoidstandup robot.
        For more information, please refer to `humanoidstandup <https://github.com/google/brax/tree/main/brax/envs/humanoidstandup.py>` env in brax.
        """
        state = super().step(state, action)

        quad_energy_cost = state.metrics["reward_quadctrl"]
        mo_reward = jnp.array([state.metrics["reward_linup"], quad_energy_cost])
        return state.replace(reward=mo_reward)



import torch


def igd(objs: torch.Tensor, pf: torch.Tensor, p: int = 1):
    """
    Calculate the Inverted Generational Distance (IGD) metric between a set of solutions and the Pareto front.

    :param objs: A tensor of shape (n, m), where n is the number of solutions and m is the number of objectives.
                 Represents the set of solutions to be evaluated.
    :param pf: A tensor of shape (k, m), where k is the number of points on the Pareto front and m is the number
               of objectives. Represents the true Pareto front.
    :param p: The power parameter used in the calculation (default is 1). This defines the distance metric (L^p norm).

    :return: The IGD score, a scalar representing the average distance of the solutions to the Pareto front.

    :note:
        The IGD score is lower when the approximation is closer to the Pareto front.
    """
    min_dis = torch.cdist(pf, objs).min(dim=1).values
    return (min_dis.pow(p).mean()).pow(1 / p)



import torch


def gd(objs: torch.Tensor, pf: torch.Tensor):
    """
    Calculate the Generational Distance (GD) metric between a set of solutions and the Pareto front.

    :param objs: A tensor of shape (n, m), where n is the number of solutions and m is the number of objectives.
                 Represents the set of solutions to be evaluated.
    :param pf: A tensor of shape (k, m), where k is the number of points on the Pareto front and m is the number
               of objectives. Represents the true Pareto front.

    :return: The GD score, a scalar representing the average distance of the solutions to the Pareto front.

    :note:
        The GD score is lower when the approximation is closer to the Pareto front.
    """
    distances = torch.cdist(objs, pf, p=2)
    min_distances = torch.min(distances, dim=1).values
    score = torch.norm(min_distances) / min_distances.size(0)
    return score



from typing import Optional

import torch
from evox.core import Problem
from evox.operators.sampling import uniform_sampling
from evox.problems.numerical.basic import griewank_func, rastrigin_func, rosenbrock_func, sphere_func

from evomo.operators.selection import non_dominate_rank


class MAF(Problem):
    """
    The MAF benchmark test suite for evolutionary many-objective optimization.

    :references:
        [1] R. Cheng, M. Li, Y. Tian, X. Zhang, S. Yang, Y. Jin, and X. Yao, "A benchmark test suite
            for evolutionary many-objective optimization," Complex & Intelligent Systems, vol. 3,
            no. 1, pp. 67-81, 2017. Available: https://link.springer.com/article/10.1007/s40747-017-0039-7
    """

    def __init__(self, d: int = None, m: int = 3, ref_num: int = 1000, device: Optional[torch.device] = None):
        """
        Initialize the MAF problem instance.

        :param d: The dimensionality of the problem (number of decision variables). If None, it is set to m + 9.
        :param m: The number of objectives for the problem. Default is 3.
        :param ref_num: The reference number for evaluating the problem, default is 1000.
        :param device: The device to perform computations on (e.g., CPU or GPU). If None, the default device is used.
        """
        super().__init__()
        self.device = device or torch.get_default_device()
        self.m = m
        self.d = self.m + 9 if d is None else d
        self.ref_num = ref_num
        self._calc_pf()

    def _calc_pf(self):
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        self._pf_value = 1 - r

    def evaluate(self, X: torch.Tensor):
        pass

    def pf(self):
        return self._pf_value


class MAF1(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        g = torch.sum(torch.pow(X[:, m - 1 :] - 0.5, 2), dim=1)

        ones_col = torch.ones(n, 1, device=X.device)
        cumprod_term = torch.flip(torch.cumprod(torch.cat([ones_col, X[:, : m - 1]], dim=1), dim=1), [1])
        reversed_term = torch.cat([ones_col, 1 - torch.flip(X[:, : m - 1], [1])], dim=1)  # Reversed slice for last term
        repeat_g = (1 + g).unsqueeze(1)

        f = repeat_g - repeat_g * cumprod_term * reversed_term
        return f


class MAF2(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        d = self.d
        g = torch.zeros(n, m, device=X.device)
        interval = int((d - m + 1) / m)
        g_cols = []
        for i in range(m):
            if i < m - 1:
                start = m + i * interval - 1
                seg = X[:, start : start + interval]
            else:
                start = m + (m - 1) * interval - 1
                seg = X[:, start:]

            seg = torch.where(seg == 0, 0.5, seg)
            g_cols.append(((seg - 0.5) ** 2).sum(dim=1, keepdim=True))
        g = torch.cat(g_cols, dim=1)

        f1 = torch.flip(
            torch.cumprod(
                torch.cat([torch.ones((n, 1), device=X.device), torch.cos((X[:, : m - 1] / 2 + 0.25) * torch.pi / 2)], dim=1),
                dim=1,
            ),
            [1],
        )
        f2 = torch.cat(
            [torch.ones(n, 1, device=X.device), torch.sin((torch.flip(X[:, : m - 1], [1]) / 2 + 0.25) * torch.pi / 2)], dim=1
        )
        f = (1 + g) * f1 * f2
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        c = torch.zeros(n, m - 1, device=r.device)

        for i in range(n):
            for j in range(1, m):
                temp = r[i, j] / r[i, 0] * torch.prod(c[i, m - j : m - 1])
                c[i, m - j - 1] = torch.sqrt(1 / (1 + temp**2))

        if m > 5:
            c = c * (
                torch.cos(torch.tensor(torch.pi / 8, device=r.device))
                - torch.cos(torch.tensor(3 * torch.pi / 8, device=r.device))
                + torch.cos(torch.tensor(3 * torch.pi / 8, device=r.device))
            )
        else:
            c = c[
                torch.all(
                    (c >= torch.cos(torch.tensor(torch.pi * 3 / 8, device=r.device)))
                    & (c <= torch.cos(torch.tensor(torch.pi / 8, device=r.device))),
                    dim=1,
                ),
                :,
            ]
        f = torch.flip(
            torch.cumprod(torch.cat([torch.ones(c.size(0), 1, device=r.device), c[:, : m - 1]], dim=1), dim=1), [1]
        ) * torch.cat([torch.ones(c.size(0), 1, device=r.device), torch.sqrt(1 - torch.flip(c[:, : m - 1], [1]) ** 2)], dim=1)
        self._pf_value = f


class MAF3(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        d = self.d
        g = 100 * (d - m + 1 + torch.sum((X[:, m - 1 :] - 0.5) ** 2 - torch.cos(20 * torch.pi * (X[:, m - 1 :] - 0.5)), dim=1))
        f1 = (
            (1 + g).unsqueeze(1)
            * torch.flip(
                torch.cumprod(
                    torch.cat([torch.ones(n, 1, device=X.device), torch.cos(X[:, : m - 1] * torch.pi / 2)], dim=1),
                    dim=1,
                ),
                [1],
            )
            * torch.cat([torch.ones(n, 1, device=X.device), torch.sin(torch.flip(X[:, : m - 1], [1]) * torch.pi / 2)], dim=1)
        )
        f = torch.cat([f1[:, : m - 1] ** 4, (f1[:, m - 1] ** 2).view(-1, 1)], dim=1)
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, self.m)

        temp = (torch.sum(torch.sqrt(r[:, :-1]), dim=1) + r[:, -1]).view(-1, 1)
        f = r / torch.cat([(temp**2).repeat(1, m - 1), temp], dim=1)
        self._pf_value = f


class MAF4(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        d = self.d
        g = 100 * (d - m + 1 + torch.sum((X[:, m - 1 :] - 0.5) ** 2 - torch.cos(20 * torch.pi * (X[:, m - 1 :] - 0.5)), dim=1))
        f1 = (1 + g).unsqueeze(1) - (1 + g).unsqueeze(1) * torch.flip(
            torch.cumprod(
                torch.cat([torch.ones(n, 1, device=X.device), torch.cos(X[:, : m - 1] * torch.pi / 2)], dim=1), dim=1
            ),
            [1],
        ) * torch.cat([torch.ones(n, 1, device=X.device), torch.sin(torch.flip(X[:, : m - 1], [1]) * torch.pi / 2)], dim=1)
        f = f1 * torch.pow(2, torch.arange(1, m + 1, device=X.device))
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        r1 = r / torch.sqrt(torch.sum(r**2, dim=1, keepdim=True))
        f = (1 - r1) * torch.pow(2, torch.arange(1, m + 1, device=r.device))
        self._pf_value = f


class MAF5(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        alpha = 100
        Xfront = X[:, : m - 1] ** alpha
        g = torch.sum((X[:, m - 1 :] - 0.5) ** 2, dim=1)
        f1 = (
            (1 + g).unsqueeze(1)
            * torch.flip(
                torch.cumprod(
                    torch.cat([torch.ones(n, 1, device=X.device), torch.cos(Xfront * torch.pi / 2)], dim=1),
                    dim=1,
                ),
                [1],
            )
            * torch.cat([torch.ones(n, 1, device=X.device), torch.sin(torch.flip(Xfront, [1]) * torch.pi / 2)], dim=1)
        )
        f = f1 * torch.pow(2, torch.arange(m, 0, -1, device=X.device))
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        r1 = r / torch.sqrt(torch.sum(r**2, dim=1, keepdim=True))
        f = r1 * torch.pow(2, torch.arange(m, 0, -1, device=r.device))
        self._pf_value = f


class MAF6(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        g = torch.sum((X[:, m - 1 :] - 0.5) ** 2, dim=1).unsqueeze(1)
        temp = g.repeat(1, m - 2)
        Xfront = X[:, : m - 1].clone()
        Xfront[:, 1:] = (1 + 2 * temp * X[:, 1 : m - 1]) / (2 + 2 * temp)
        f = (
            (1 + 100 * g)
            * torch.flip(
                torch.cumprod(
                    torch.cat([torch.ones(n, 1, device=X.device), torch.cos(Xfront * torch.pi / 2)], dim=1),
                    dim=1,
                ),
                [1],
            )
            * torch.cat([torch.ones(n, 1, device=X.device), torch.sin(torch.flip(Xfront, [1]) * torch.pi / 2)], dim=1)
        )
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, 2)
        r1 = r / torch.sqrt(torch.sum(r**2, dim=1, keepdim=True))

        if r1.size(1) < m:
            r1 = torch.cat((r1[:, torch.zeros(m - r1.size(1), device=r.device, dtype=torch.long)], r1), dim=1)
        f = r1 / torch.pow(torch.sqrt(torch.tensor(2.0)), torch.maximum(torch.tensor(self.m - 2), torch.tensor(0)).repeat(n, 1))

        self._pf_value = f


class MAF7(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 19, f"{self.__class__.__name__} is only defined for d = m + 19, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        g = 1 + 9 * torch.mean(X[:, m - 1 :], dim=1)
        fFront = X[:, : m - 1]
        fRear = (1 + g) * (
            m
            - torch.sum(
                fFront / (1 + g.unsqueeze(1)) * (1 + torch.sin(3 * torch.pi * fFront)),
                dim=1,
            )
        )
        f = torch.cat([fFront, fRear.unsqueeze(1)], dim=1)
        return f

    @torch.jit.ignore
    def _calc_pf(self):
        m = self.m
        interval = torch.tensor([0, 0.251412, 0.631627, 0.859401], device=self.device)
        median = (interval[1] - interval[0]) / (interval[3] - interval[2] + interval[1] - interval[0])
        X = self._grid(self.ref_num * self.m, m - 1)
        X = torch.where(X <= median, X * (interval[1] - interval[0]) / median + interval[0], X)
        X = torch.where(
            X > median,
            (X - median) * (interval[3] - interval[2]) / (1 - median) + interval[2],
            X,
        )
        f = torch.cat(
            [
                X,
                2 * (self.m - torch.sum(X / 2 * (1 + torch.sin(3 * torch.pi * X)), dim=1)).view(-1, 1),
            ],
            dim=1,
        )
        self._pf_value = f

    def _grid(self, N: int, M: int):
        gap = torch.linspace(0, 1, steps=int(N ** (1 / M)), device=self.device)
        mesh = torch.meshgrid(*([gap] * M), indexing="ij")
        W = torch.cat([x.reshape(-1, 1) for x in mesh], dim=1)
        return W


class MAF8(MAF):
    def __init__(self, d=2, m=10, ref_num=1000, device: Optional[torch.device] = None):
        assert d == 2, f"{self.__class__.__name__} is only defined for d = 2, got {d}."
        assert m >= 3, f"{self.__class__.__name__} is only defined for m >= 3, got {m}."
        super().__init__(d, m, ref_num, device)
        self.points = self._getPoints()

    def evaluate(self, X: torch.Tensor):
        f = self._eucl_dis(X[:, :2], self.points)
        return f

    def _calc_pf(self):
        if not hasattr(self, "points"):
            self.points = self._getPoints()

        temp = torch.linspace(
            -1, 1, steps=int(torch.sqrt(torch.tensor(self.ref_num * self.m, device=self.device))), device=self.device
        )
        y, x = torch.meshgrid(temp, temp, indexing="ij")
        x = x.flatten()
        y = y.flatten()
        _points = torch.stack([x, y], dim=-1)
        ND = torch.stack([self._point_in_polygon(self.points, p) for p in _points])
        f = self._eucl_dis(torch.stack([x[ND], y[ND]], dim=-1), self.points)
        self._pf_value = f

    def _eucl_dis(self, X: torch.Tensor, Y: torch.Tensor):
        return torch.cdist(X, Y)

    def _getPoints(self):
        m = self.m
        theta, rho = self._cart2pol(torch.tensor(0, device=self.device), torch.tensor(1, device=self.device))
        temp = torch.arange(1, m + 1, device=self.device).view(-1, 1)
        x, y = self._pol2cart(theta - temp * 2 * torch.pi / m, rho)
        return torch.cat([x, y], dim=1)

    def _cart2pol(self, x: torch.Tensor, y: torch.Tensor):
        rho = torch.sqrt(x**2 + y**2)
        theta = torch.atan2(y, x)
        return theta, rho

    def _pol2cart(self, theta: torch.Tensor, rho: torch.Tensor):
        x = rho * torch.cos(theta)
        y = rho * torch.sin(theta)
        return x, y

    def _inside(self, x, a, b):
        return (torch.minimum(a, b) <= x) & (x < torch.maximum(a, b))

    def _ray_intersect_segment(self, point, seg_init, seg_term):
        y_dist = seg_term[1] - seg_init[1]
        # special case: y_dist == 0, check P_y == seg_init_y and P_x inside the segment
        judge_1 = (point[1] == seg_init[1]) & self._inside(point[0], seg_init[0], seg_term[0])
        # check intersection_x >= P_x.
        LHS = seg_init[0] * y_dist + (point[1] - seg_init[1]) * (seg_term[0] - seg_init[0])
        RHS = point[0] * y_dist
        # since it's an inequation, reverse the inequation if y_dist is negative.
        judge_2 = ((y_dist > 0) & (LHS >= RHS)) | ((y_dist < 0) & (LHS <= RHS))
        # check intersection_y, which is P_y is inside the segment
        judge_3 = self._inside(point[1], seg_init[1], seg_term[1])
        return ((y_dist == 0) & judge_1) | ((y_dist != 0) & judge_2 & judge_3)

    def _point_in_polygon(self, polygon, point):
        seg_term = torch.roll(polygon, 1, dims=0)
        is_intersect_list = []

        for i in range(polygon.size(0)):
            intersect = self._ray_intersect_segment(point, polygon[i], seg_term[i])
            is_intersect_list.append(intersect.unsqueeze(0))

        is_intersect = torch.cat(is_intersect_list, dim=0)
        is_vertex = torch.any(torch.all(polygon == point, dim=1))
        return (torch.sum(is_intersect) % 2 == 1) | is_vertex


class MAF9(MAF8):
    def __init__(self, d=2, m=10, ref_num=1000, device: Optional[torch.device] = None):
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        f_cols = []
        for i in range(self.points.size(0) - 1):
            col_i = self._Point2Line(X, self.points[i : i + 2, :])
            f_cols.append(col_i.unsqueeze(1))
        last_col = self._Point2Line(X, torch.cat([self.points[-1, :].unsqueeze(0), self.points[0, :].unsqueeze(0)], dim=0))
        f_cols.append(last_col.unsqueeze(1))

        f = torch.cat(f_cols, dim=1)
        return f

    def _Point2Line(self, pop_dec: torch.Tensor, line: torch.Tensor):
        Distance = torch.abs(
            (line[0, 0] - pop_dec[:, 0]) * (line[1, 1] - pop_dec[:, 1])
            - (line[1, 0] - pop_dec[:, 0]) * (line[0, 1] - pop_dec[:, 1])
        ) / torch.sqrt((line[0, 0] - line[1, 0]) ** 2 + (line[0, 1] - line[1, 1]) ** 2)
        return Distance


class MAF10(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        d = self.d
        s = torch.arange(2, 2 * m + 1, step=2, device=X.device)
        z01 = X / torch.arange(2, d * 2 + 1, step=2, device=X.device)
        t0Front = z01[:, : m - 1]
        t0Rear = self._s_linear(z01[:, m - 1 :], 0.35)
        t0 = torch.cat([t0Front, t0Rear], dim=1)

        t = self._evaluate(t0, X)

        tRear = t[:, m - 1 : m]
        xFront = torch.maximum(tRear, torch.ones_like(tRear)) * (t[:, : m - 1] - 0.5) + 0.5
        x = torch.cat([xFront, tRear], dim=1)

        h_convex = self._convex(x)[:, : m - 1]
        h_mixed = self._mixed(x).unsqueeze(1)
        h = torch.cat([h_convex, h_mixed], dim=1)

        f = tRear + s * h
        return f

    def _calc_pf(self):
        m = self.m
        x, temp, a = self._pf_a()
        e = torch.abs(
            temp.unsqueeze(1) * (1 - torch.cos(torch.pi / 2 * a))
            - 1
            + (a + torch.cos(10 * torch.pi * a + torch.pi / 2) / 10 / torch.pi)
        )
        rank = torch.argsort(e, dim=1)

        x[:, 0] = a[0, torch.min(rank[:, :10], dim=1)[1]]
        f = self._convex(x)
        f[:, m - 1] = self._mixed(x)
        f = f * torch.arange(2, 2 * m + 1, 2, device=self.device)
        self._pf_value = f

    def _evaluate(self, t1: torch.Tensor, X: torch.Tensor):
        m = self.m
        d = self.d
        t2 = t1.clone()
        t2[:, m - 1 :] = self._b_flat(t1[:, m - 1 :], 0.8, 0.75, 0.85)

        t3 = t2**0.02

        outs = []
        for i in range(m - 1):
            y = t3[:, i : i + 2]
            w = torch.arange(2 * i, 2 * (i + 1) + 1, 2, device=X.device)
            outs.append(self._r_sum(y, w))

        y_last = t3[:, m - 1 : d]
        w_last = torch.arange(2 * m, 2 * d + 1, 2, device=X.device)
        outs.append(self._r_sum(y_last, w_last))

        t = torch.stack(outs, dim=1)

        return t

    def _s_linear(self, Y: torch.Tensor, a):
        return torch.abs(Y - a) / torch.abs(torch.floor(a - Y) + a)

    def _b_flat(self, Y: torch.Tensor, a, b, c):
        output = (
            a
            + torch.minimum(torch.zeros_like(Y - b, device=Y.device), torch.floor(Y - b)) * a * (b - Y) / b
            - torch.minimum(torch.zeros_like(c - Y, device=Y.device), torch.floor(c - Y)) * (1 - a) * (Y - c) / (1 - c)
        )
        return torch.round(output * 1e4) / 1e4

    def _r_sum(self, Y: torch.Tensor, W: torch.Tensor):
        return torch.sum(Y * W, dim=1) / torch.sum(W)

    def _convex(self, x: torch.Tensor):
        ones = torch.ones(x.size(0), 1, device=x.device)
        return torch.flip(
            torch.cumprod(torch.cat([ones, 1 - torch.cos(x[:, :-1] * torch.pi / 2)], dim=1), dim=1), [1]
        ) * torch.cat([ones, 1 - torch.sin(torch.flip(x[:, :-1], [1]) * torch.pi / 2)], dim=1)

    def _mixed(self, x: torch.Tensor):
        return 1 - x[:, 0] - torch.cos(10 * torch.pi * x[:, 0] + torch.pi / 2) / (10 * torch.pi)

    def _pf_a(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        c = torch.ones(n, m, device=r.device)

        for i in range(n):
            for j in range(1, m):
                temp = r[i, j] / r[i, 0] * torch.prod(1 - c[i, m - j : m - 1])
                c[i, m - j - 1] = (temp**2 - temp + torch.sqrt(2 * temp)) / (temp**2 + 1)

        x = torch.arccos(c) * 2 / torch.pi
        temp = (1 - torch.sin(torch.pi / 2 * x[:, 1])) * r[:, m - 1] / r[:, m - 2]
        a = torch.arange(0, 1.0001, 0.0001, device=r.device).unsqueeze(0)
        return x, temp, a


class MAF11(MAF10):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        d = m + 9 if d is None else d
        d = int((d - m + 1) / 2) * 2 + m - 1
        super().__init__(d, m, ref_num, device)

    def _calc_pf(self):
        m = self.m
        x, temp, a = self._pf_a()
        e = torch.abs(temp.unsqueeze(1) * (1 - torch.cos(torch.pi / 2 * a)) - 1 + a * torch.cos(5 * torch.pi * a) ** 2)
        rank = torch.argsort(e, dim=1)
        x[:, 0] = a[0, torch.min(rank[:, :10], dim=1)[1]]
        f = self._convex(x)
        f[:, m - 1] = self._mixed(x)
        non_dominated_rank = non_dominate_rank(f)
        f = f[non_dominated_rank == 0, :]
        f = f * torch.arange(2, 2 * m + 1, 2, device=x.device)
        self._pf_value = f

    def _evaluate(self, t1: torch.Tensor, X: torch.Tensor):
        m = self.m
        d = self.d
        L = d - (m - 1)

        t2Front = t1[:, : m - 1]
        t2Rear = (t1[:, m - 1 :: 2] + t1[:, m::2] + 2 * torch.abs(t1[:, m - 1 :: 2] - t1[:, m::2])) / 3
        t2 = torch.cat([t2Front, t2Rear], dim=1)

        outs = []
        w = torch.ones(1, device=X.device)
        for i in range(m - 1):
            y = t2[:, i : i + 1]
            outs.append(self._r_sum(y, w))

        y_last = t2[:, m - 1 : m - 1 + L // 2]
        w_last = torch.ones(L // 2, device=X.device)
        outs.append(self._r_sum(y_last, w_last))

        t = torch.stack(outs, dim=1)
        return t

    def _mixed(self, x: torch.Tensor):
        return self._disc(x)

    def _disc(self, x):
        return 1 - x[:, 0] * (torch.cos(5 * torch.pi * x[:, 0])) ** 2


class MAF12(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == m + 9, f"{self.__class__.__name__} is only defined for d = m + 9, got {d}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        d = self.d
        L = d - (m - 1)
        S = torch.arange(2, 2 * m + 1, step=2, device=X.device)

        z01 = X / torch.arange(2, d * 2 + 1, step=2, device=X.device)

        Y = (torch.flip(torch.cumsum(torch.flip(z01, [1]), dim=1), [1]) - z01) / torch.arange(d - 1, -1, -1, device=X.device)
        t1Front = z01[:, : d - 1] ** (
            0.02
            + (50 - 0.02)
            * (0.98 / 49.98 - (1 - 2 * Y[:, : d - 1]) * torch.abs(torch.floor(0.5 - Y[:, : d - 1]) + 0.98 / 49.98))
        )
        t1Rear = z01[:, d - 1 :]
        t1 = torch.cat([t1Front, t1Rear], dim=1)

        t2Front = self._s_decept(t1[:, : m - 1], 0.35, 0.001, 0.05)
        t2Rear = self._s_multi(t1[:, m - 1 :], 30, 95, 0.35)
        t2 = torch.cat([t2Front, t2Rear], dim=1)

        t3_cols = []
        for i in range(m - 1):
            temp = t2[:, i : i + 1]
            t3_cols.append(self._r_nonsep(temp, 1))

        diff = torch.abs(t2Rear.unsqueeze(2) - t2Rear.unsqueeze(1))
        SUM = diff.sum(dim=(1, 2)) * 0.5

        denom1 = torch.tensor(L / 2, device=X.device)
        denom2 = 1 + 2 * L - 2 * denom1
        last_col = (t2Rear.sum(dim=1) + 2 * SUM) / denom1 / denom2
        t3_cols.append(last_col)

        t3 = torch.stack(t3_cols, dim=1)

        x_cols = []
        max_factor = torch.maximum(last_col, torch.ones_like(last_col))
        for i in range(m - 1):
            x_cols.append(max_factor * (t3[:, i] - 0.5) + 0.5)
        x_cols.append(last_col)
        x = torch.stack(x_cols, dim=1)
        h = self._concave(x)
        f = x[:, m - 1].unsqueeze(1) + S * h
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        r = r / torch.sqrt(torch.sum(r**2, dim=1)).reshape(-1, 1)
        f = torch.arange(2, 2 * m + 1, 2) * r
        self._pf_value = f

    def _s_decept(self, Y: torch.Tensor, a, b, c):
        return 1 + (torch.abs(Y - a) - b) * (
            torch.floor(Y - a + b) * (1 - c + (a - b) / b) / (a - b)
            + torch.floor(a + b - Y) * (1 - c + (1 - a - b) / b) / (1 - a - b)
            + 1 / b
        )

    def _s_multi(self, Y: torch.Tensor, a, b, c):
        return (
            1
            + torch.cos((4 * a + 2) * torch.pi * (0.5 - torch.abs(Y - c) / 2 / (torch.floor(c - Y) + c)))
            + 4 * b * (torch.abs(Y - c) / 2 / (torch.floor(c - Y) + c)) ** 2
        ) / (b + 2)

    def _r_nonsep(self, Y: torch.Tensor, a):
        s1 = Y.sum(dim=1)
        if a > 1:
            s2 = sum((Y - Y.roll(-o, dims=1)).abs().sum(dim=1) for o in range(1, a))
        else:
            s2 = torch.zeros_like(s1)
        int_a2 = a // 2

        return (s1 + s2) / (Y.size(1) / a) * int_a2 * (1 + 2 * a - 2 * int_a2)

    def _concave(self, X: torch.Tensor):
        return torch.flip(
            torch.cumprod(
                torch.cat([torch.ones(X.shape[0], 1, device=X.device), torch.sin(X[:, :-1] * torch.pi / 2)], dim=1),
                dim=1,
            ),
            [1],
        ) * torch.cat(
            [
                torch.ones(X.size(0), 1, device=X.device),
                torch.cos(torch.flip(X[:, :-1], [1]) * torch.pi / 2),
            ],
            dim=1,
        )


class MAF13(MAF):
    def __init__(self, d=5, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert m >= 3, f"{self.__class__.__name__} is only defined for m >= 3, got {m}."
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        d = self.d
        Y = X - 2 * X[:, 1].view(-1, 1) * torch.sin(
            2 * torch.pi * X[:, 0].view(-1, 1) + torch.arange(1, d + 1, device=X.device) * torch.pi / d
        )
        f = torch.zeros(n, m, device=X.device)
        f0 = torch.sin(X[:, 0] * torch.pi / 2) + 2 * torch.mean(Y[:, 3:d:3] ** 2, dim=1)
        f1 = torch.cos(X[:, 0] * torch.pi / 2) * torch.sin(X[:, 1] * torch.pi / 2) + 2 * torch.mean(Y[:, 4:d:3] ** 2, dim=1)
        f2 = torch.cos(X[:, 0] * torch.pi / 2) * torch.cos(X[:, 1] * torch.pi / 2) + 2 * torch.mean(Y[:, 2:d:3] ** 2, dim=1)
        f3 = (
            (f[:, 0] ** 2 + f[:, 1] ** 10 + f[:, 2] ** 10 + 2 * torch.mean(Y[:, 3:d] ** 2, dim=1))
            .unsqueeze(1)
            .repeat(1, self.m - 3)
        )
        f = torch.cat([f0.unsqueeze(1), f1.unsqueeze(1), f2.unsqueeze(1), f3], dim=1)
        return f

    def _calc_pf(self):
        m = self.m
        r, n = uniform_sampling(self.ref_num * self.m, 3)
        r = r / torch.sqrt(torch.sum(r**2, dim=1, keepdim=True))
        f = torch.cat([r, (r[:, 0] ** 2 + r[:, 1] ** 10 + r[:, 2] ** 10).unsqueeze(1).repeat(1, m - 3)], dim=1)
        self._pf_value = f


class MAF14(MAF):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        assert d == 20 * m, f"{self.__class__.__name__} is only defined for d = 20 * m, got {d}."
        super().__init__(d, m, ref_num, device)
        nk = 2
        c = torch.zeros(self.m, device=device)
        c[0] = 3.8 * 0.1 * (1 - 0.1)
        for i in range(1, self.m):
            c[i] = 3.8 * c[i - 1] * (1 - c[i - 1])

        self.sublen = torch.floor(c / torch.sum(c) * (self.d - self.m + 1) / nk)
        self.len = torch.cat([torch.tensor([0]), torch.cumsum(self.sublen * nk, dim=0)], dim=0)
        self.sublen = tuple(map(int, self.sublen))
        self.len = tuple(map(int, self.len))

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        g = self._evaluate(X)
        f = (
            (1 + g)
            * torch.flip(torch.cumprod(torch.cat([torch.ones(n, 1, device=X.device), X[:, : m - 1]], dim=1), dim=1), [1])
            * torch.cat([torch.ones(n, 1, device=X.device), 1 - torch.flip(X[:, : m - 1], [1])], dim=1)
        )
        return f

    def _calc_pf(self):
        self._pf_value = uniform_sampling(self.ref_num * self.m, self.m)[0]

    def _evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        nk = 2
        new_X = self._modify_X(X)
        g = torch.zeros(n, m, device=X.device)

        for i in range(0, m, 2):
            g = self._inner_loop(i, self._func1, g, nk, new_X)

        for i in range(1, m, 2):
            g = self._inner_loop(i, self._func2, g, nk, new_X)
        g = g / torch.tensor(self.sublen, device=X.device).unsqueeze(0) * nk
        return g

    def _modify_X(self, X: torch.Tensor):
        new_X = X.clone()
        new_X[:, self.m - 1 :] = (1 + torch.arange(self.m, self.d + 1, device=X.device) / self.d).unsqueeze(0) * X[
            :, self.m - 1 :
        ] - (X[:, 0] * 10).unsqueeze(-1)
        return new_X

    def _inner_loop(self, i, inner_fun, g: torch.Tensor, nk, X: torch.Tensor):
        new_col = g[:, i].clone()
        for j in range(1, nk):
            start = self.len[i] + self.m - 1 + j * self.sublen[i]
            end = start + self.sublen[i]
            temp = X[:, start:end]
            new_col = new_col + inner_fun(temp)
        g_out = torch.cat(
            [g[:, :i], new_col.unsqueeze(1), g[:, i + 1 :]],
            dim=1,
        )
        return g_out

    def _func1(self, X):
        return rastrigin_func(X)

    def _func2(self, X):
        return rosenbrock_func(X)


class MAF15(MAF14):
    def __init__(self, d=None, m=3, ref_num=1000, device: Optional[torch.device] = None):
        super().__init__(d, m, ref_num, device)

    def evaluate(self, X: torch.Tensor):
        m = self.m
        n = X.size(0)
        g = self._evaluate(X)
        f = (1 + g + torch.cat([g[:, 1:], torch.zeros(n, 1, device=X.device)], dim=1)) * (
            1
            - torch.flip(
                torch.cumprod(
                    torch.cat([torch.ones(n, 1, device=X.device), torch.cos(X[:, : m - 1] * torch.pi / 2)], dim=1),
                    dim=1,
                ),
                [1],
            )
            * torch.cat([torch.ones(n, 1, device=X.device), torch.sin(torch.flip(X[:, : m - 1], [1]) * torch.pi / 2)], dim=1)
        )
        return f

    def _calc_pf(self):
        r, n = uniform_sampling(self.ref_num * self.m, self.m)
        r = 1 - r / torch.sqrt(torch.sum(r**2, axis=1)).reshape(-1, 1)
        self._pf_value = r

    def _modify_X(self, X: torch.Tensor):
        new_X = X.clone()
        new_X[:, self.m - 1 :] = (
            1 + torch.cos(torch.arange(self.m, self.d + 1, device=X.device).unsqueeze(0) / self.d * torch.pi / 2)
        ) * X[:, self.m - 1 :] - (X[:, 0] * 10).unsqueeze(-1)
        return new_X

    def _func1(self, X):
        return griewank_func(X)

    def _func2(self, X):
        return sphere_func(X)



import math
from typing import Callable, Optional

import torch
from evox.core import Algorithm, Mutable, vmap
from evox.operators.crossover import simulated_binary_half
from evox.operators.mutation import polynomial_mutation
from evox.operators.sampling import uniform_sampling
from evox.utils import clamp


def pbi(f, w, z):
    norm_w = torch.norm(w, dim=1)
    f = f - z
    d1 = torch.sum(f * w, dim=1) / norm_w
    d2 = torch.norm(f - (d1[:, None] * w / norm_w[:, None]), dim=1)
    return d1 + 5 * d2


def tchebycheff(f, w, z):
    return torch.max(torch.abs(f - z) * w, dim=1)[0]


def tchebycheff_norm(f, w, z, z_max):
    return torch.max(torch.abs(f - z) / (z_max - z) * w, dim=1)[0]


def modified_tchebycheff(f, w, z):
    return torch.max(torch.abs(f - z) / w, dim=1)[0]


def weighted_sum(f, w):
    return torch.sum(f * w, dim=1)


def shuffle_rows(matrix: torch.Tensor) -> torch.Tensor:
    """
    Shuffle each row of the given matrix independently without using a for loop.

    Args:
        matrix (torch.Tensor): A 2D tensor.

    Returns:
        torch.Tensor: A new tensor with each row shuffled differently.
    """
    rows, cols = matrix.size()

    permutations = torch.argsort(torch.rand(rows, cols, device=matrix.device), dim=1)
    return matrix.gather(1, permutations)

class TensorMOEAD(Algorithm):
    """
    TensorMOEA/D

    This is a tensorized implementation of the original MOEA/D algorithm, which incorporates GPU acceleration
    for improved computational performance in solving multi-objective optimization problems.

    :references:
        [1] Q. Zhang and H. Li, "MOEA/D: A Multiobjective Evolutionary Algorithm Based on Decomposition,"
            IEEE Transactions on Evolutionary Computation, vol. 11, no. 6, pp. 712-731, 2007. Available:
            https://ieeexplore.ieee.org/document/4358754

        [2] Z. Liang, H. Li, N. Yu, K. Sun, and R. Cheng, "Bridging Evolutionary Multiobjective Optimization and
            GPU Acceleration via Tensorization," IEEE Transactions on Evolutionary Computation, 2025. Available:
            https://ieeexplore.ieee.org/document/10944658

    :note: This implementation differs from the original MOEA/D algorithm by incorporating tensorization for
            GPU acceleration, significantly improving performance for large-scale optimization tasks.
    """

    def __init__(
        self,
        pop_size: int,
        n_objs: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        aggregate_op=("pbi", "pbi"),
        selection_op: Optional[Callable] = None,
        mutation_op: Optional[Callable] = None,
        crossover_op: Optional[Callable] = None,
        device: torch.device | None = None,
    ):
        """Initializes the TensorMOEA/D algorithm.

        :param pop_size: The size of the population.
        :param n_objs: The number of objective functions in the optimization problem.
        :param lb: The lower bounds for the decision variables (1D tensor).
        :param ub: The upper bounds for the decision variables (1D tensor).
        :param aggregate_op: The aggregation function to use for the algorithm (optional).
        :param selection_op: The selection operation for evolutionary strategy (optional).
        :param mutation_op: The mutation operation, defaults to `polynomial_mutation` if not provided (optional).
        :param crossover_op: The crossover operation, defaults to `simulated_binary` if not provided (optional).
        :param device: The device on which computations should run (optional). Defaults to PyTorch's default device.
        """

        super().__init__()
        self.pop_size = pop_size
        self.n_objs = n_objs
        device = torch.get_default_device() if device is None else device
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.shape[0]
        # write to self
        self.lb = lb.to(device=device)
        self.ub = ub.to(device=device)

        self.selection = selection_op
        self.mutation = mutation_op
        self.crossover = crossover_op

        if self.mutation is None:
            self.mutation = polynomial_mutation
        if self.crossover is None:
            self.crossover = simulated_binary_half

        w, _ = uniform_sampling(self.pop_size, self.n_objs)
        w = w.to(device=device)

        self.pop_size = w.size(0)
        assert self.pop_size > 10, "Population size must be greater than 10. Please reset the population size."
        self.n_neighbor = int(math.ceil(self.pop_size / 10))

        length = ub - lb
        population = torch.rand(self.pop_size, self.dim, device=device)
        population = length * population + lb

        neighbors = torch.cdist(w, w)
        self.neighbors = torch.argsort(neighbors, dim=1, stable=True)[:, : self.n_neighbor]
        self.w = w

        self.pop = Mutable(population)
        self.fit = Mutable(torch.full((self.pop_size, self.n_objs), torch.inf, device=device))
        self.z = Mutable(torch.zeros((self.n_objs,), device=device))

        self.aggregate_func1 = self.get_aggregation_function(aggregate_op[0])
        self.aggregate_func2 = self.get_aggregation_function(aggregate_op[1])

    def get_aggregation_function(self, name: str) -> Callable:
        aggregation_functions = {
            "pbi": pbi,
            "tchebycheff": tchebycheff,
            "tchebycheff_norm": tchebycheff_norm,
            "modified_tchebycheff": modified_tchebycheff,
            "weighted_sum": weighted_sum,
        }
        if name not in aggregation_functions:
            raise ValueError(f"Unsupported function: {name}")
        return aggregation_functions[name]

    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.fit = self.evaluate(self.pop)
        self.z = torch.min(self.fit, dim=0)[0]

    def step(self):
        """Perform the optimization step of the workflow."""
        parent = shuffle_rows(self.neighbors)
        selected_p = torch.cat([self.pop[parent[:, 0]], self.pop[parent[:, 1]]], dim=0)

        crossovered = self.crossover(selected_p)
        offspring = self.mutation(crossovered, self.lb, self.ub)
        offspring = clamp(offspring, self.lb, self.ub)
        off_fit = self.evaluate(offspring)

        self.z = torch.min(self.z, torch.min(off_fit, dim=0)[0])

        sub_pop_indices = torch.arange(0, self.pop_size, device=self.pop.device)
        update_mask = torch.zeros((self.pop_size,), dtype=torch.bool, device=self.pop.device)

        def body(ind_p, ind_obj):
            g_old = self.aggregate_func1(self.fit[ind_p], self.w[ind_p], self.z)
            g_new = self.aggregate_func1(ind_obj, self.w[ind_p], self.z)
            temp_mask = update_mask.clone()
            temp_mask = torch.scatter(temp_mask, 0, ind_p, g_old > g_new)
            return torch.where(temp_mask, -1, sub_pop_indices.clone())

        replace_indices = vmap(body, in_dims=(0, 0))(self.neighbors, off_fit)

        def update_population(sub_indices, population, pop_obj, w_ind):
            f = torch.where(sub_indices[:, None] == -1, off_fit, pop_obj)
            x = torch.where(sub_indices[:, None] == -1, offspring, population)
            idx = torch.argmin(self.aggregate_func2(f, w_ind[None, :], self.z))
            return x[idx], f[idx]

        self.pop, self.fit = vmap(update_population, in_dims=(1, 0, 0, 0))(replace_indices, self.pop, self.fit, self.w)



from functools import partial

import torch
from evox.core import Problem


def _generic_zdt(f1, g, h, x):
    f1_x = f1(x)
    g_x = g(x)
    return torch.stack([f1_x, g_x * h(f1_x, g_x)],dim=1)


class ZDT(Problem):
    """
    The ZDT test suite for multiobjective optimization.

    This suite of benchmark functions is commonly used to evaluate the performance of multiobjective
    evolutionary algorithms.

    :references:
        [1] E. Zitzler, K. Deb, and L. Thiele, "Comparison of Multiobjective Evolutionary Algorithms:
            Empirical Results," in Evolutionary Computation, vol. 8, no. 2, pp. 173-195, 2000.
            Available: https://ieeexplore.ieee.org/abstract/document/6787994
    """

    def __init__(self, n: int, ref_num: int = 100):
        """
        Initialize the ZDT problem instance.

        :param n: The dimensionality of the problem (number of variables).
        :param ref_num: The reference number for evaluating the problem, default is 100.
        """
        super().__init__()
        self.n = n
        self._zdt = None
        self.ref_num = ref_num

    def evaluate(self, X: torch.Tensor):
        return self._zdt(X)

    def pf(self):
        x = torch.linspace(0, 1, self.ref_num)
        return torch.stack([x, 1 - torch.sqrt(x)], dim=1)


class ZDT1(ZDT):
    def __init__(self, n):
        super().__init__(n)
        def f1(x):
            return x[:,0]
        def g(x):
            return 1 + 9 * torch.mean(x[:,1:])
        def h(f1, g):
            return 1 - torch.sqrt(f1 / g)
        self._zdt = partial(_generic_zdt, f1, g, h)


class ZDT2(ZDT):
    def __init__(self, n):
        super().__init__(n)
        def f1(x):
            return x[:,0]
        def g(x):
            return 1 + 9 * torch.mean(x[:,1:])
        def h(f1_val, g_val):
            return 1 - (f1_val / g_val) ** 2
        self._zdt = partial(_generic_zdt, f1, g, h)

    def pf(self):
        x = torch.linspace(0, 1, self.ref_num)
        return torch.stack([x, 1 - x**2], dim=1)


class ZDT3(ZDT):
    def __init__(self, n):
        super().__init__(n)
        def f1(x):
            return x[:,0]
        def g(x):
            return 1 + 9 * torch.mean(x[:,1:])
        def h(f1, g):
            return 1 - torch.sqrt(f1 / g) - (f1 / g) * torch.sin(10 * torch.pi * f1)
        self._zdt = partial(_generic_zdt, f1, g, h)

    def pf(self):
        r = torch.tensor(
            [
                [0.0000, 0.0830],
                [0.1822, 0.2577],
                [0.4093, 0.4538],
                [0.6183, 0.6525],
                [0.8233, 0.8518],
            ]
        )

        pf_points = []
        segment_size = self.ref_num // len(r)
        for row in r:
            x_vals = torch.linspace(row[0].item(), row[1].item(), segment_size)
            f2_vals = 1 - torch.sqrt(x_vals) - x_vals * torch.sin(10 * torch.pi * x_vals)
            pf_points.append(torch.stack([x_vals, f2_vals], dim=1))
        pf = torch.cat(pf_points, dim=0)
        return pf


class ZDT4(ZDT):
    def __init__(self, n):
        super().__init__(n)
        def f1(x):
            return x[:,0]
        def g(x):
            return 1 + 10 * (self.n - 1) + torch.sum(x[:,1:] ** 2 - 10.0 * torch.cos(4.0 * torch.pi * x[:,1:]))
        def h(f1_val, g_val):
            return 1 - torch.sqrt(f1_val / g_val)
        self._zdt = partial(_generic_zdt, f1, g, h)


class ZDT6(ZDT):
    def __init__(self, n):
        super().__init__(n)
        def f1(x):
            return 1 - torch.exp(-4.0 * x[:,0]) * torch.sin(6.0 * torch.pi * x[:,0]) ** 6
        def g(x):
            return 1 + 9.0 * (torch.sum(x[:,1:]) / 9.0) ** 0.25
        def h(f1_val, g_val):
            return 1 - (f1_val / g_val) ** 2
        self._zdt = partial(_generic_zdt, f1, g, h)

    def pf(self):
        min_f1 = 0.280775
        f1_vals = torch.linspace(min_f1, 1.0, self.ref_num)
        f2_vals = 1.0 - f1_vals**2
        return torch.stack([f1_vals, f2_vals], dim=1)



import jax.numpy as jnp
from brax.envs.hopper import Hopper


class MoHopper(Hopper):
    """
    The multi-objective version of hopper env (2 objectives).

    :references:
        [1] C. D. Freeman, E. Frey, A. Raichuk, S. Girgin, I. Mordatch, and O. Bachem,
        “Brax - a differentiable physics engine for large scale rigid body simulation,” 2021.
        [Online]. Available: http://github.com/google/brax
    """
    def __init__(self, **kwargs):
        """Initialize the multi-objective hopper env.

        :param num_obj: The number of the objectives. For this env, it is set to 2.
        """
        super().__init__(**kwargs)
        self.num_obj = 2

    def reset(self, rng):
        state = super().reset(rng)
        mo_reward = jnp.zeros((self.num_obj, ))
        return state.replace(reward=mo_reward)

    def step(self, state, action):
        """Run one timestep of the environment's dynamics.

        :param height: The height at which the hopper robot is located.
        For more information, please refer to `hopper <https://github.com/google/brax/tree/main/brax/envs/hopper.py>` env in brax.
        """
        state = super().step(state, action)
        init_z = self.sys.link.transform.pos[0, 2]
        z = state.pipeline_state.x.pos[0, 2]
        height = 10 * (z - init_z)
        mo_reward = jnp.array([state.metrics['reward_forward'], height])
        mo_reward += state.metrics['reward_ctrl']
        mo_reward += state.metrics['reward_healthy']
        return state.replace(reward=mo_reward)



from typing import Callable, Optional

import torch
from evox.core import Algorithm, Mutable
from evox.operators.crossover import simulated_binary
from evox.operators.mutation import polynomial_mutation
from evox.operators.selection import tournament_selection
from evox.utils import clamp, lexsort

from evomo.operators.selection import non_dominate_rank


def cal_hv(fit: torch.Tensor, ref: torch.Tensor, pop_size: int, n_sample: int):
    n, m = fit.size()
    alpha = torch.cumprod(torch.cat([torch.ones(1, device=fit.device), (pop_size - torch.arange(1, n, device=fit.device)) / (n - torch.arange(1, n, device=fit.device))]), dim=0) / torch.arange(1, n + 1, device=fit.device)
    alpha = torch.nan_to_num(alpha)

    f_min = torch.min(fit, dim=0).values

    samples = torch.rand(n_sample, m, device=fit.device) * (ref - f_min) + f_min

    ds = torch.zeros(n_sample, dtype=torch.int64, device=fit.device)
    pds = (fit.unsqueeze(0).expand(n_sample, -1, -1) - samples.unsqueeze(1).expand(-1, n, -1) <= 0).all(dim=2)
    ds = torch.sum(torch.where(pds, ds.unsqueeze(1) + 1, ds.unsqueeze(1)), dim=1)
    ds = torch.where(ds == 0, ds, ds - 1)

    temp = torch.where(pds.T, ds.unsqueeze(0), -1)
    value = torch.where(temp != -1, alpha[temp], torch.tensor(0, dtype=torch.float32))
    f = torch.sum(value, dim=1)

    f = f * torch.prod(ref - f_min) / n_sample
    return f


class HypE(Algorithm):
    """
    The tensorized version of HypE algorithm.

    :references:
        [1] J. Bader and E. Zitzler, "HypE: An algorithm for fast hypervolume-based many-objective optimization,"
            Evolutionary Computation, vol. 19, no. 1, pp. 45-76, 2011. Available:
            https://direct.mit.edu/evco/article-abstract/19/1/45/1363/HypE-An-Algorithm-for-Fast-Hypervolume-Based-Many

        [2] Z. Liang, H. Li, N. Yu, K. Sun, and R. Cheng, "Bridging Evolutionary Multiobjective Optimization and
            GPU Acceleration via Tensorization," IEEE Transactions on Evolutionary Computation, 2025. Available:
            https://ieeexplore.ieee.org/document/10944658
    """

    def __init__(
        self,
        pop_size: int,
        n_objs: int,
        lb: torch.Tensor,
        ub: torch.Tensor,
        n_sample: int = 10000,
        selection_op: Optional[Callable] = None,
        mutation_op: Optional[Callable] = None,
        crossover_op: Optional[Callable] = None,
        device: torch.device | None = None,
    ):
        """Initializes the HypE algorithm.

        :param pop_size: The size of the population.
        :param n_objs: The number of objective functions in the optimization problem.
        :param lb: The lower bounds for the decision variables (1D tensor).
        :param ub: The upper bounds for the decision variables (1D tensor).
        :param n_sample: The number of samples for hypervolume calculation (optional).
        :param selection_op: The selection operation for evolutionary strategy (optional).
        :param mutation_op: The mutation operation, defaults to `polynomial_mutation` if not provided (optional).
        :param crossover_op: The crossover operation, defaults to `simulated_binary` if not provided (optional).
        :param device: The device on which computations should run (optional). Defaults to PyTorch's default device.
        """

        super().__init__()
        self.pop_size = pop_size
        self.n_objs = n_objs
        if device is None:
            device = torch.get_default_device()
        # check
        assert lb.shape == ub.shape and lb.ndim == 1 and ub.ndim == 1
        assert lb.dtype == ub.dtype and lb.device == ub.device
        self.dim = lb.shape[0]
        # write to self
        self.lb = lb.to(device=device)
        self.ub = ub.to(device=device)
        self.n_sample = n_sample

        self.selection = selection_op
        self.mutation = mutation_op
        self.crossover = crossover_op

        self.selection = tournament_selection
        if self.mutation is None:
            self.mutation = polynomial_mutation
        if self.crossover is None:
            self.crossover = simulated_binary

        length = ub - lb
        population = torch.rand(self.pop_size, self.dim, device=device)
        population = length * population + lb

        self.ref = Mutable(torch.ones(n_objs, device=device))

        self.pop = Mutable(population)
        self.fit = Mutable(torch.full((self.pop_size, self.n_objs), torch.inf, device=device))


    def init_step(self):
        """
        Perform the initialization step of the workflow.

        Calls the `init_step` of the algorithm if overwritten; otherwise, its `step` method will be invoked.
        """
        self.fit = self.evaluate(self.pop)
        self.ref = torch.full((self.n_objs,), torch.max(self.fit).item() * 1.2, device=self.fit.device)

    def step(self):
        """Perform the optimization step of the workflow."""
        hv = cal_hv(self.fit, self.ref, self.pop_size, self.n_sample)
        mating_pool = self.selection(self.pop_size, -hv)
        crossovered = self.crossover(self.pop[mating_pool])
        offspring = self.mutation(crossovered, self.lb, self.ub)
        offspring = clamp(offspring, self.lb, self.ub)
        off_fit = self.evaluate(offspring)

        merge_pop = torch.cat([self.pop, offspring], dim=0)
        merge_fit = torch.cat([self.fit, off_fit], dim=0)

        rank = non_dominate_rank(merge_fit)
        order = torch.argsort(rank)
        worst_rank = rank[order[self.pop_size - 1]]
        mask = rank <= worst_rank

        hv = cal_hv(merge_fit, self.ref, torch.sum(mask) - self.pop_size, self.n_sample)
        dis = torch.where(mask, hv, -torch.inf)

        combined_indices = lexsort([-dis, rank])[: self.pop_size]

        self.pop = merge_pop[combined_indices]
        self.fit = merge_fit[combined_indices]
